"""
Broadcasting support
"""

from __future__ import annotations

from moat.util import NotGiven, Queue
from moat.lib.micro import EndOfStream, WouldBlock

from typing import TYPE_CHECKING, Generic, TypeVar, cast

if TYPE_CHECKING:
    from typing import Self

try:
    from weakref import WeakSet
except ImportError:
    WeakSet = set  # type:ignore[assignment]

TData = TypeVar("TData")


# TODO build something nicer
try:
    import logging
except ImportError:
    logger = None
else:
    logger = logging.getLogger(__name__)

__all__ = [
    "BroadcastReader",
    "Broadcaster",
    "LostData",
]


class LostData(Exception):
    """
    Indicator of data loss.

    Attribute ``n`` contains the number of dropped messages.
    """

    def __init__(self, n: int) -> None:
        self.n = n


class BroadcastReader(Generic[TData]):
    """
    The read side of a broadcaster.

    Simply iterate over it.

    Warning: The iterator may raise ``LostData`` exceptions.
    These contain the number of messages that have been dropped
    due to the reader being too slow.

    Call this object to inject a value.
    """

    def __init__(
        self,
        parent: Broadcaster,
        length: int = 1,
        skip: bool = False,
    ) -> None:
        self.parent = parent
        self.length = length
        self.skip = skip
        self.value: TData = NotGiven
        self.loss: int = 0

        if self.length <= 0:
            raise RuntimeError("Length must be at least one")
        self._q: Queue = Queue(self.length)

    def __hash__(self) -> int:
        return id(self)

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> TData:
        if not self.skip and self.loss > 0:
            n, self.loss = self.loss, 0
            raise LostData(n)

        try:
            return await self._q.get()
        except (AttributeError, EndOfStream, EOFError):
            raise StopAsyncIteration from None

    def __enter__(self):
        return self

    async def __aenter__(self):
        return self

    def __exit__(self, *tb) -> None:
        self.close()

    async def __aexit__(self, *tb) -> None:
        self.close()

    def flush(self) -> None:
        """
        Clean the queue.

        Useful for re-sync after you get a `LostData` error.
        """
        try:
            while True:
                self._q.get_nowait()
        except WouldBlock:
            return

    def __call__(self, value: TData) -> None:
        """enqueue a value, to this reader only"""
        try:
            self._q.put_nowait(value)
        except WouldBlock:
            x = self._q.get_nowait()
            if logger is not None:
                logger.debug("Dropped: %r", x)
            self._q.put_nowait(value)
            self.loss += 1

    def _close(self) -> None:
        self._q.close_sender()

    def close(self) -> None:
        "close this reader, detaching it from its parent"
        self._close()
        self.parent._closed_reader(self)  # noqa:SLF001 pylint: disable=protected-access

    async def aclose(self) -> None:
        "close this reader, detaching it from its parent"
        self.close()


class Broadcaster(Generic[TData]):
    """
    A simple broadcaster. Messages will be sent to all readers.

    Args:
        length:
            readers' default queue length.
        send_last:
            if set, a new reader immediately gets the last-sent value,
            if any. Otherwise it strictly waits for new data.

    If a queue is full, the oldest message will be discarded. Readers will
    then get a LostData exception that contains the number of dropped
    messages.

    Writing requires a context manager (sync or async).
    Then simply call with a value.

    ::

        async with anyio.create_task_group() as tg, Broadcaster() as bc:
            tg.start_soon(rdr, aiter(bc))
            for x in range(5):
                bc(x)
                await anyio.sleep(0.01)
            bc(42)

    To read, async-iterate::

        async def rdr(bcr: Broadcaster):
            async with bcr.reader() as mq:
                async for msg in mq:
                    print(msg)

    alternately::

        async def rdr(bcr: BroadcastReader|Broadcaster):
            async with aclosing(aiter(bcr)) as mq:
                async for msg in mq:
                    print(msg)

    Do **not** skip the *aclosing*, esp. on MicroPython.

    To safely re-sync, do something like this::

        bcr.flush()
        x = fetch_consistent_state()  # this really should be sync
        bcr(x)  # if this fails, you've got a problem
        while True:
            y = await anext(bcr)
            if x is y:
                break

    """

    def __init__(
        self,
        length: int = 1,
        send_last: bool = False,
    ) -> None:
        self.length = length
        self.send_last = send_last
        self._rdr: WeakSet[BroadcastReader] | None = None
        self.value: TData = NotGiven

    def open(self) -> Self:
        """Open the broadcaster.

        Consider using a context instead of this method.
        """
        if self._rdr is not None:
            raise RuntimeError("already entered/opened")
        self._rdr = WeakSet()
        return self

    def __enter__(self) -> Self:
        return self.open()

    async def __aenter__(self) -> Self:
        return self.open()

    def __exit__(self, *tb) -> None:
        self.close()

    async def __aexit__(self, *tb) -> None:
        self.close()

    def _closed_reader(self, reader: BroadcastReader[TData]) -> None:
        if self._rdr is not None:
            self._rdr.remove(reader)

    def __aiter__(self) -> BroadcastReader[TData]:
        """Create a reader with the predefined queue length"""
        assert self._rdr is not None

        r: BroadcastReader[TData] = BroadcastReader(self, self.length)
        self._rdr.add(r)
        if self.send_last and self.value is not NotGiven:
            r(cast("TData", self.value))
        return aiter(r)

    def reader(
        self, length: int, send_last: bool | None = None, skip: bool = False
    ) -> BroadcastReader[TData]:
        """Create a reader with an explicit queue length.

        Args:
            length: Queue size.
            skip: if `True`, ignore overflows.
        """
        assert self._rdr is not None

        if send_last is None:
            send_last = self.send_last
        r: BroadcastReader[TData] = BroadcastReader(self, length, skip)
        self._rdr.add(r)
        if send_last and self.value is not NotGiven:
            if not length:
                raise ValueError("This would deadlock. Use length>0.")
            r(cast("TData", self.value))
        return r

    def __call__(self, value: TData) -> None:
        """Enqueue a value to all readers"""
        assert self._rdr is not None

        self.value = value
        for r in self._rdr:
            r(value)

    async def read(self) -> TData:
        "gets the last value (waits until there is one)"
        if self.value is NotGiven:
            return await anext(aiter(self))
        return self.value

    def close(self) -> None:
        "Close the broadcaster. No more writing."
        if self._rdr is not None:
            for r in self._rdr:
                r._close()  # noqa:SLF001 pylint: disable=protected-access
            self._rdr = None



"""
Module for pins
"""

from __future__ import annotations

import asyncio
import machine as M

from moat.util import NotGiven
from moat.lib.micro import AC_use, Event, TaskGroup
from moat.lib.rpc import BaseCmd


class _Pin:
    """
    A config-enabled pin that you can async-iterate for changes.

        p = Pin(attrdict(pin=3), mode=machine.Pin.IN)
        with p:
            async for val in p:
                print("Pin",p,"is now",val)

    All other import arguments are taken from keywords.
    """

    def __init__(self, *a, **kw):
        self._pin = M.Pin(*a, **kw)
        self.flag = asyncio.ThreadSafeFlag()
        self.evt = Event()
        self.val = self._pin.value()

    def _irq(self, _x=None):
        "sets the change flag"
        self.val = self._pin.value()
        self.flag.set()

    def __call__(self, n=None):
        if n is None:
            return self._pin.value()
        else:
            self._pin.value(n)
            self.val = self._pin.value()
            self.flag.set()

    async def flag_watch(self):
        "Flag reader task, since a ThreadSafeFlag only acepts one read task"
        try:
            self._pin.irq(self._irq, M.Pin.IRQ_FALLING | M.Pin.IRQ_RISING)
            while True:
                await self.flag.wait()
                self.flag.clear()
                self.evt.set()
                self.evt = Event()
        finally:
            self._pin.irq(None)

    async def __aenter__(self):
        self.flag.set()

    async def __aexit__(self, *err):
        pass

    def __aiter__(self):
        return self

    async def __anext__(self):
        await self.evt.wait()
        return self.pin()


class Pin(BaseCmd):
    """
    This is a basic digital pin.

    Iterating it yields a new value whenever the pin changes.

    Parameters:
        out(bool): direction (input: `False`)
        drive(int): drive strength. ``0…3`` on ESP32.
        init(bool): initial value. Leave alone if not given.
        pull(bool|None): Pull-up (`True`), -down (`False`).
        open(bool|None): open-collector/drain (`True`) or
                         -emitter/source (`False`), on output.
    """

    doc = dict(
        _c=dict(
            _d="Digital I/O pin",
            pin="int:Nr",
            out="bool:output?",
            init="bool:initial out state",
            drive="int:strength",
            pull="bool|None: Pullup/down?",
            open="bool|None: open-collector/emitter?",
        ),
        _d="r/w",
        _a=[
            dict(_r="bool:current state"),
            dict(
                _0="bool:new state",
                o="bool:old: wait until pin value differs",
            ),
        ],
        _o=True,
        _s=True,
    )

    def __init__(self, cfg):
        super().__init__(cfg)
        out = cfg.get("out", False)
        oce = cfg.get("open", None)  # open collector/emitter

        a = [cfg["pin"], (M.Pin.OPEN_DRAIN if oce else M.Pin.OUT) if out else M.Pin.IN]
        kw = {}
        if (val := cfg.get("init", None)) is not None:
            kw["value"] = val
        if (drive := cfg.get("drive", None)) is not None:
            kw["drive"] = getattr(M.Pin, f"DRIVE_{drive}", 0)
        if (pull := cfg.get("pull", None)) is not None:
            a.append(M.Pin.PULL_UP if pull else M.Pin.PULL_DOWN)

        self.pin = _Pin(*a, **kw)

    async def setup(self):
        "initialization, triggers change"
        await super().setup()
        if getattr(self, "tg", None) is None:
            self.tg = await AC_use(self, TaskGroup())
        await self.tg.spawn(self.pin.flag_watch)

    async def cmd(self, val: bool = NotGiven) -> None | bool:
        "Simple Data protocol."
        if val is NotGiven:
            return self.pin()
        self.pin(val)

    async def stream(self, msg):
        """
        R/W data stream. The initial argument says whether you want to
        read. Waits for change if @o (old value) is not None.
        """
        o = msg.get("o", None)
        async with TaskGroup() as tg, msg.stream() as m:

            async def _rd():
                val = bool(self.pin())
                if o is not val:
                    await m.send(val)
                while True:
                    await self.pin.evt.wait()
                    await m.send(bool(self.pin()))

            if o is not None or msg.get(0, False):
                tg.start_soon(_rd)

            for mm in m:
                self.pin(mm[0])
            tg.cancel()
            await msg.result()



"""
This is a (temporary) standalone main for the Gaggenau controller replacement.
"""

from __future__ import annotations

import os
import sys
from time import sleep_ms, ticks_diff, ticks_ms, ticks_us

try:
    from micropython import const
except ImportError:

    def const(x):
        "Fake const"
        return x


import network
from mplex import Multiplex

from machine import Pin

from moat.micro.rtc import RTC

# Network configuration


cfg = dict(
    wlan=dict(
        SSID="***",
        PASSWORD="*** *** ***",  # noqa:S106
        STATIC_IP="10.107.0.42",
        SUBNET="255.255.255.0",
        GATEWAY="10.107.0.1",
        DNS="8.8.8.8",
    ),
    light=21,  # inverted
    triac=dict(
        input=0,
        output=20,  # inverted
    ),
    buttons=dict(
        scan=[22, 23, 3, 2],
        input=[10, 11],
        output=[12, 13],
        inv=dict(
            scan=True,
            input=False,
            output=False,
        ),
        msec=5,
        timer=1,
    ),
)

wlan = None


def wifi(cfg: dict | None = None):
    "setup wifi"
    global wlan

    if wlan is None:
        wlan = network.WLAN(network.STA_IF)
    if cfg is None:
        wlan.active(False)
        return

    wlan.active(True)

    # Configure static IP
    wlan.ifconfig((cfg["STATIC_IP"], cfg["SUBNET"], cfg["GATEWAY"], cfg["DNS"]))
    wlan.connect(cfg["SSID"], cfg["PASSWORD"])


def mwifi():
    "manually open the wifi link"
    wifi(cfg["wlan"])
    import webrepl  # noqa:PLC0415

    webrepl.start()


B_Fn = const(0)
B_Light = const(1)
B_Time = const(2)
B_Int = const(3)
B_3 = const(4)
B_2 = const(5)
B_1 = const(6)
B_0 = const(7)


def rtc(name, default):
    "read rtc state"
    return RTC.get_sync(name, default=default)


class Gaggenau:
    "Controller"

    wifi_do = False
    t_wifi_last = None
    t_wifi_ok = None
    mp: Multiplex | None = None

    def __init__(self, cfg):
        self.cfg = cfg
        self.mp = Multiplex(cfg["buttons"])
        self._light = Pin(cfg["light"], Pin.OUT, value=1)
        self._fan = Pin(cfg["triac"]["output"], Pin.OUT, value=1)

    def start(self):
        "Setup"
        self.mp.start()
        self._fan(not rtc("fan", 0))
        self._light(not rtc("light", 0))
        self.t_wifi_last = ticks_ms()

        if rtc("wifi", False):
            RTC.set_sync("wifi", False)
            self.wifi_do = True
            wifi(True)

        if self.fan:
            self.mp[B_3] = True
        if self.light:
            self.mp[B_Light] = True

    def stop(self):
        "Teardown"
        self.mp.stop()
        self._fan(1)
        self._light(1)

    @property
    def light(self):
        "Light state"
        return not self._light()

    @light.setter
    def light(self, val):
        "Light state"
        self._light(not val)
        RTC.set_sync("light", val)

    @property
    def fan(self):
        "Fan state"
        return not self._fan()

    @fan.setter
    def fan(self, val):
        "Fan state"
        self._fan(not val)
        RTC.set_sync("fan", val)

    def run(self):
        "Run loop"
        try:
            print("* MoaT Herd * Task starts *")
            self.start()
            while True:
                try:
                    self.step()
                except Exception as exc:
                    sys.print_exception(exc)
                    sleep_ms(1000)
                else:
                    sleep_ms(100)
        except BaseException:
            self.stop()
            raise

    def time_step(self):
        "time a step"
        a = ticks_us()
        self.step()
        b = ticks_us()
        return ticks_diff(b, a)

    def step(self):
        "Run one pass"
        tm = ticks_ms()
        for btn, val in self.mp.changes():
            print(btn, val)
            if not val:
                return
            if btn == B_Light:
                self.light = not self.light
                self.mp[B_Light] = self.light
            elif btn == B_3:
                self.fan = 1
                self.mp[B_1] = 0
                self.mp[B_2] = 0
                self.mp[B_3] = 1
            elif btn == B_0:
                self.fan = 0
                self.mp[B_1] = 0
                self.mp[B_2] = 0
                self.mp[B_3] = 0
            elif btn == B_Fn:
                if rtc("wifi", False) or self.wifi_do or (wlan is not None and wlan.isconnected()):
                    wifi(None)
                    self.wifi_do = False
                    self.mp[B_Fn] = False
                    RTC.set_sync("wifi", False)
                    self.t_wifi_last = None
                else:
                    wifi(self.cfg["wlan"])
                    self.mp[B_Fn] = True
                    self.t_wifi_last = ticks_ms()
                    self.wifi_do = True

        # Now handle wifi.
        if self.t_wifi_last is not None and ticks_diff(tm, self.t_wifi_last) > 750:
            self.t_wifi_last = tm
            if self.wifi_do:
                if wlan.isconnected():
                    self.wifi_do = False
                    self.t_wifi_ok = ticks_ms()
                    self.mp[B_Fn] = True

                    try:
                        os.stat("webrepl_cfg.py")
                    except OSError:
                        pass
                    else:
                        import webrepl  # noqa:PLC0415

                        webrepl.start()
                else:
                    self.mp[B_Fn] = not self.mp.get_out(B_Fn)

            elif wlan is None or not wlan.isconnected():
                self.mp[B_Fn] = False

            else:
                self.mp[B_Fn] = True
                if self.t_wifi_ok is not None and ticks_diff(tm, self.t_wifi_last) > 30000:
                    self.t_wifi_ok = None
                    RTC.set_sync("wifi", True)


if rtc("run", True):
    ga = Gaggenau(cfg)
    ga.run()



"""
Loopback helpers for RPC streams.
"""

from __future__ import annotations

from moat.lib.micro import log, shield
from moat.util.exc import ungroup

from .const import B_FLAGSTR
from .stream.base import HandlerStream, wire2i_f

try:
    import anyio
except ImportError:
    import asyncio

    def cancelled_class():
        return asyncio.CancelledError

else:
    cancelled_class = anyio.get_cancelled_exc_class

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import MsgRoot

__all__ = ["StreamLoop"]


class StreamLoop(HandlerStream):
    "A test stream that implements loopback"

    __other: StreamLoop | None = None

    def __init__(self, h: MsgRoot, s: str):
        super().__init__(h)
        self.__s = s

    def attach_remote(self, other):
        "attach the other end to us"
        self.__other = other

    async def write_stream(self):
        "write loop"
        other = self.__other
        if other is None:
            raise RuntimeError("No remote")
        while True:
            try:
                msg = await self.msg_out()
            except EOFError:
                return
            m = msg[:]
            i, fl = wire2i_f(m.pop(0))
            f = B_FLAGSTR[fl]
            if i >= 0:
                f += "+"
            f += str(i)

            log("%s: %s %s", self.__s, f, " ".join(repr(x) for x in m))
            await other.msg_in(msg)

    async def read_stream(self):
        "read loop. No-op; all work is done by the other side's writer."
        other = self.__other
        if other is None:
            raise RuntimeError("No remote")
        await other.writer_done.wait()

    async def __aexit__(self, *tb):
        other = self.__other
        if other is None:
            raise RuntimeError("No remote")
        with shield():
            await other.closed_input()
        try:
            with ungroup:
                await super().__aexit__(*tb)
        finally:
            if not self.is_idle:
                log("*** WARNING *** %r: not idle; %r", self, vars(self))
            # assert self.is_idle

        if isinstance(ungroup.one(tb[1]), cancelled_class()):
            return True



"""
Utility functions for serial port apps.
"""

from __future__ import annotations

from moat.micro.part.serial import NamedSerial


def get_serial(cfg):
    """
    Get the appropriate serial class for the given config.

    Args:
        cfg: Configuration dict with 'port' key.

    Returns:
        A serial port instance.
    """
    p = cfg["port"]
    if not isinstance(p, str):
        Ser = p
    elif p == "USB":
        Ser = getattr(
            __import__("moat.micro.part.serial", fromlist=("USBSerial",)),
            "USBSerial",
            None,
        )
        if Ser is None:
            Ser = NamedSerial
    else:
        Ser = NamedSerial
    return Ser(cfg)



"""
Stream link-up support for MoaT commands
"""

from __future__ import annotations

from moat.lib.micro import ACM, AC_exit, L, Lock, TaskGroup
from moat.lib.rpc import BaseCmd
from moat.lib.stream import BaseBlk, BaseBuf, BaseMsg

__all__ = ["BaseCmdBBM"]

# Typing
from typing import TYPE_CHECKING, cast  # isort:skip

if TYPE_CHECKING:
    from moat.lib.stream.base import Buffer, MutBuffer

    from collections.abc import Awaitable
    from typing import Protocol

    class _ConsoleProto(Protocol):
        async def crd(self, buf: MutBuffer) -> int: ...

        async def cwr(self, buf: Buffer) -> None: ...


class BaseCmdBBM(BaseCmd):
    """
    This is a command handler that connects MoaT's Cmd tree to a
    `~moat.lib.stream.BaseBuf`, `~moat.lib.stream.BaseBlk` or
    `~moat.lib.stream.BaseMsg` instance.

    Override :meth:`stream` to return that instance. Use
    :func:`moat.lib.micro.AC_use` if it's a context manager.

    This is a single class that adapts `~moat.lib.stream.BaseMsg`,
    `~moat.lib.stream.BaseBlk`, and `~moat.lib.stream.BaseBuf` streams.

    The difference between this and a
    :moat.lib.rpc.cmd.msg:`BaseCmdMsg`-derived class is that this
    class exposes commands that directly access the underlying stream
    (of whatever type).

    In contrast, :moat.lib.rpc.cmd.msg:`BaseCmdMsg` encapsulates
    arbitrary commands and requires a
    :moat.lib.rpc.cmd.msg:`BaseCmdMsg` handler on the other side to
    talk to.

    This class cannot wrap a pre-existing stream, by design.
    """

    s = None

    def __init__(self, cfg):
        super().__init__(cfg)
        self.w_lock = Lock()

    async def stream(self) -> BaseMsg | BaseBlk | BaseBuf:
        "create the actual data stream. Override this!"
        raise NotImplementedError("setup", self.path)

    async def setup(self):  # noqa:D102
        await super().setup()
        self.s = await self.stream()

    async def run(self):  # noqa:D102
        ACM(self)
        try:
            await super().run()
        finally:
            self.s = None
            await AC_exit(self)

    # Buf: rd/wr = .rd/.wr

    doc_rd = dict(_d="read bytestream", _0="int:len (64)")

    async def cmd_rd(self, n=64) -> Buffer:
        """read some data"""
        if L:
            await self.wait_ready()
        b = bytearray(n)
        if self.s is None:
            raise EOFError
        r = await cast(BaseBuf, self.s).rd(b)
        if r == n:
            return b
        elif r <= n >> 2:
            return bytes(b[:r])
        else:
            b = memoryview(b)
            return b[:r]

    doc_wr = dict(_d="write bytestream", _0="bytes:data")

    async def cmd_wr(self, b: Buffer):
        """write some data"""
        if L:
            await self.wait_ready()
        async with self.w_lock:
            if self.s is None:
                raise EOFError
            await cast(BaseBuf, self.s).wr(b)

    doc_rw = dict(
        _d="r/w byte stream", _0="int:rdbuflen (64)", _i="bytes:to send", _o="bytes:received"
    )

    async def stream_rw(self, msg):
        "read/write byte stream"
        n = msg.get(0, 64)
        async with msg.stream() as st, TaskGroup() as tg:

            @tg.start_soon
            async def rd():
                while True:
                    await st.send(await self.cmd_rd(n))

            async for m in st:
                await self.cmd_wr(m[0])
            tg.cancel()

    # Blk/Msg: Console crd/cwr = .crd/cwr

    doc_crd = dict(_d="read console", _0="int:len (64)", _r="bytes:data")

    async def cmd_crd(self, n=64) -> Buffer:
        """read some console data"""
        b = bytearray(n)
        if self.s is None:
            raise EOFError
        r = await cast("_ConsoleProto", self.s).crd(b)
        if r == n:
            return b
        elif r <= n >> 2:
            return bytes(b[:r])
        else:
            b = memoryview(b)
            return b[:r]

    doc_cwr = dict(_d="write console", _0="bytes:data")

    async def cmd_cwr(self, b: Buffer):
        """write some console data"""
        async with self.w_lock:
            if self.s is None:
                raise EOFError
            await cast("_ConsoleProto", self.s).cwr(b)

    doc_crw = dict(
        _d="r/w console stream", _0="int:rdbuflen (64)", _i="bytes:to send", _o="bytes:received"
    )

    async def stream_crw(self, msg):
        "read/write console stream"
        n = msg.get(0, 64)
        async with msg.stream() as st, TaskGroup() as tg:

            @tg.start_soon
            async def crd():
                while True:
                    await st.send(await self.cmd_crd(n))

            async for m in st:
                await self.cmd_cwr(m[0])
            tg.cancel()

    # Msg: s/r = .send/.recv

    doc_s = dict(_d="write message", _0="any:message")

    def cmd_s(self, m) -> Awaitable:  # pylint:disable=invalid-overridden-method
        """send a message"""
        if self.s is None:
            raise EOFError
        return cast(BaseMsg, self.s).send(m)

    doc_r = dict(_d="read message", _r="any:message")

    def cmd_r(self) -> Awaitable:  # pylint:disable=invalid-overridden-method
        """receive a message"""
        if self.s is None:
            raise EOFError
        return cast(BaseMsg, self.s).recv()

    doc_mrw = dict(_d="r/w msg stream", _i="any:message", _o="any:message")

    async def stream_mrw(self, msg):
        "read/write message stream"
        async with msg.stream() as st, TaskGroup() as tg:

            @tg.start_soon
            async def mrd():
                while True:
                    await st.send(await self.cmd_r())

            async for m in st:
                await self.cmd_s(m[0])
            tg.cancel()

    # Blk: sb/rb = .snd/.rcv

    doc_sb = dict(_d="write block", _0="bytes:encoded message")

    def cmd_sb(self, m) -> Awaitable:  # pylint:disable=invalid-overridden-method
        """send a binary message"""
        if self.s is None:
            raise EOFError
        return cast(BaseBlk, self.s).snd(m)

    doc_rb = dict(_d="read block", _r="any:encoded message")

    def cmd_rb(self) -> Awaitable:  # pylint:disable=invalid-overridden-method
        """receive a binary message"""
        if self.s is None:
            raise EOFError
        return cast(BaseBlk, self.s).rcv()

    doc_brw = dict(_d="r/w block stream", _i="bytes:message", _o="bytes:message")

    async def stream_brw(self, msg):
        "read/write block stream"
        async with msg.stream() as st, TaskGroup() as tg:

            @tg.start_soon
            async def brd():
                while True:
                    await st.send(await self.cmd_rb())

            async for m in st:
                await self.cmd_sb(m[0])
            tg.cancel()



"""
Docstrings for mode and serial
"""

from __future__ import annotations

_mode_d = dict(
    baudrate="int:baud (9600)",
    parity="bool|None:odd?",
    flow="str:RC for hardware",
    stopbits="int:1 or 2",
    bytesize="int: bits (8)",
    rts_state="bool:RTS on?",
    dtr_state="bool:DTR on?",
    rts_flip="bool:RTS flip on open?",
    dtr_flip="bool:DTR flip on open?",
    delay="float:wait after open",
    dtr_rts="float:wait between dtr and rts flip",
    delay_flip="float:wait after flip",
    flush="float|bool:flush inbuf after open (s, True=0.2)",
)
_cons_d = "bool|int:process non-framed data (default no; int=buflen)"
_frame_d = dict(
    max_idle="int:time between frames(ms)",
    max_packet="int:max len (127)",
    frame_start="int:start byte (xFA)",
    mark="int:each-framed-char marker (None)",
)
_loss_d = dict(
    _x="bool:enable w/ defaults",
    window="int:msgs in transit (8)",
    timeout="int:retransmit (ms)",
    retries="int:repeats until error(5)",
)
_link_d = dict(console=_cons_d, frame=_frame_d, lossy=_loss_d)

_log_d = dict(log="str:log highlevel", log_rel="str:log lossy HL", log_raw="str:log lowlevel")



"""
This module implements the direct connection to a micropython board.

MoaT uses this to can sync the initial files and get things running.
"""

from __future__ import annotations

import anyio
import ast
import logging
import os
import re
from anyio.streams.buffered import BufferedByteReceiveStream

from moat.lib.micro import AC_use
from moat.lib.stream import SingleAnyioBuf

from .os_error_list import os_error_mapping

# Typing
from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from anyio.abc import ByteStream


logger = logging.getLogger(__name__)

re_oserror = re.compile(r"OSError: (\[Errno )?(\d+)(\] )?")
re_exceptions = re.compile(r"(ValueError|KeyError|ImportError): (.*)")


async def _noop_hook(ser):
    pass


class DirectREPL(SingleAnyioBuf):
    """
    Interface to the remote REPL
    """

    serial: ByteStream | None = None
    srbuf: BufferedByteReceiveStream | None = None

    def _serial(self) -> ByteStream:
        ser = self.serial
        if ser is None:
            raise RuntimeError("serial stream not initialized")
        return ser

    def _srbuf(self) -> BufferedByteReceiveStream:
        buf = self.srbuf
        if buf is None:
            raise RuntimeError("buffered stream not initialized")
        return buf

    async def stream(self):
        "Context. Tries hard to exit special MicroPython modes, if any"
        ser = await super().stream()
        if not hasattr(ser, "send") or not hasattr(ser, "receive"):
            raise TypeError(f"Not a serial stream: {ser!r}")
        self.serial = ser

        async def out_():
            ser = self._serial()
            with anyio.fail_after(0.5):
                await ser.send(b"\x02\x03\x03")

        await AC_use(self, out_)

        ser = self._serial()
        self.srbuf = BufferedByteReceiveStream(ser)

        await ser.send(b"\x02")  # exit raw repl (if we're in that)
        await self.flush_in(0.5)
        with anyio.fail_after(0.5):
            await ser.send(b"\x03")  # CTRL+C (if running something)
        await self.flush_in(0.5)
        with anyio.fail_after(0.5):
            await ser.send(b"\x01")  # enter raw repl
        await self.flush_in(0.2)

        # Rather than wait for a longer timeout we try sending a command.
        # Most likely the first time will go splat because the response
        # doesn't start with "OK", but that's fine, just try again.
        try:
            await self.exec_raw("print(1)")
        except (OSError, TimeoutError):
            await ser.send(b"\x02\x03")  # exit raw repl, CTRL+C
            await self.flush_in(0.2)
            await ser.send(b"\x03\x01")  # CTRL+C, enter raw repl
            try:
                await anyio.sleep(0.2)
                await self.exec_raw("print(2)")
            except OSError:
                await anyio.sleep(0.2)
                await self.exec_raw("print(3)")
        return ser

    async def flush_in(self, timeout=0.1):
        "flush incoming data"
        started = False
        b = b""
        ser = self._serial()
        while True:
            with anyio.move_on_after(timeout):
                res = await ser.receive(200)
                if not started:
                    logger.debug("Flushing…")
                    started = True
                b = b + res  # [-200:]
                continue
            break
        if b:
            logger.debug("Flush: IN\n%s", b.decode("utf-8"))
        self._srbuf()._buffer = bytearray()  # noqa:SLF001 pylint: disable=protected-access

    def _parse_error(self, text):
        """Read the error message and convert exceptions"""
        lines = text.splitlines()
        if lines[0].startswith("Traceback"):
            m = re_oserror.match(lines[-1])
            if m:
                err_num = int(m.group(2))
                if err_num == 2:
                    raise FileNotFoundError(2, "File not found")
                if err_num == 13:
                    raise PermissionError(13, "Permission Error")
                if err_num == 17:
                    raise FileExistsError(17, "File Already Exists Error")
                if err_num == 19:
                    raise OSError(err_num, "No Such Device Error")
                if err_num:
                    raise OSError(err_num, os_error_mapping.get(err_num, (None, "OSError"))[1])
            m = re_exceptions.match(lines[-1])
            if m:
                raise __builtins__[m.group(1)](m.group(2))

    async def exec_raw(self, cmd, timeout=5, quiet=False):
        """Exec code, returning (stdout, stderr)"""
        if not quiet:
            logger.debug("Exec: %r", cmd)
        ser = self._serial()
        srbuf = self._srbuf()
        await ser.send(cmd.encode("utf-8"))
        await ser.send(b"\x04")

        if not timeout:
            logger.debug("does not return")
            return "", ""  # dummy output if timeout=0 was specified

        try:
            with anyio.fail_after(timeout):
                data = await srbuf.receive_until(b"\x04>", max_bytes=10000)
        except TimeoutError:
            # interrupt, read output again to get the expected traceback message
            logger.debug("Timeout. Buffer:\n%s\n", srbuf.buffer)
            await ser.send(b"\x03")  # CTRL+C
            with anyio.fail_after(3):
                data = await srbuf.receive_until(b"\x04>", max_bytes=10000)

        try:
            out, err = data.split(b"\x04")
        except ValueError:
            raise OSError(f"CTRL-D missing in response: {data!r}") from None

        if b"\nStart MoaT:" in out:
            i = out.find(b"\nOK")
            if i > 0:
                out = out[i + 1 :]
        if not out.startswith(b"OK"):
            raise OSError(f"data was not accepted: {out}: {err}")
        out = out[2:].decode("utf-8")
        err = err.decode("utf-8")
        if not quiet:
            if out:
                logger.debug("OUT %r", out)
            if err:
                logger.debug("ERR %r", err)
        return out, err

    async def exec(self, cmd, timeout=3, quiet=False):
        """run a command"""
        if not cmd.endswith("\n"):
            cmd += "\n"
        out, err = await self.exec_raw(cmd, timeout, quiet=quiet)
        if err:
            self._parse_error(err)
            raise OSError(f"execution failed: {out}: {err}")
        return out

    async def evaluate(self, cmd, quiet=False):
        """
        :param str code: code to execute
        :returns: Python object

        Execute the string (using :meth:`exec`) and return the output
        parsed using ``ast.literal_eval`` so that numbers, strings, lists etc.
        can be handled as Python objects.
        """
        return ast.literal_eval(await self.exec(cmd, quiet=quiet))

    async def reset(self):
        """
        Send a hard reset command
        """
        await self.exec_raw("import machine; machine.reset()", timeout=0, quiet=False)

    async def soft_reset(self, run_main: bool = True):
        """
        :param bool run_main: select if program should be started

        Perform a soft reset of the target. ``main.py`` will not be
        executed if ``run_main`` is True (the default).
        """
        if run_main:
            # exit raw REPL for a reset that runs main.py
            await self._serial().send(b"\x03\x03\x02\x04\x01")
        else:
            # if raw REPL is active, then MicroPython will not execute main.py
            await self._serial().send(b"\x03\x03\x04\x01")
            # execute empty line to get a new prompt
            # and consume all the outputs form the soft reset
            try:
                await anyio.sleep(0.1)
                await self.exec("4")
            except OSError:
                try:
                    await anyio.sleep(0.2)
                    await self.exec("5")
                except OSError:
                    await anyio.sleep(0.2)
                    await self.exec("6")

    async def statvfs(self, path: str) -> os.statvfs_result:
        """
        Return statvfs information (disk size, free space etc.) about a remote
        filesystem.
        """
        st = await self.evaluate(f"import os; print(os.statvfs({str(path)!r}))")
        return os.statvfs_result(st)
        # ~ f_bsize, f_frsize, f_blocks, f_bfree, f_bavail,
        #   f_files, f_ffree, f_favail, f_flag, f_namemax

    async def truncate(self, path: str, length: int):
        """
        Truncate a file.

        MicroPython has no ``file.truncate()``, but ``open(...,"ab"); write(b"")``
        seems to work.
        """
        return await self.evaluate(
            f'_f = open({str(path)!r}, "ab")\n'
            f"print(_f.seek({int(length)}))\n"
            '_f.write(b"")\n'
            "_f.close(); del _f",
        )



"""
A rudimentary debugger for MicroPython.
"""

from __future__ import annotations

import gc
import sys

# Typing

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from typing import ClassVar


# ruff: noqa:D101,D102,D103


class MpdbQuit(Exception):
    pass


class Mbdb:
    def __init__(self):
        self.botframe = None
        self.breaks = {}
        self.quitting = False
        self.currentbp = None
        self.stopframe = None
        self.curframe = None
        self.stoplineno = -1

    def reset(self):
        self.botframe = None
        self._set_stopinfo(None, None)
        self.curframe = None

    def run(self, cmd):  # noqa:ARG002
        self.reset()

    def trace_dispatch(self, frame, event, arg):
        self.curframe = frame
        if self.quitting:
            return  # None
        if event == "line":
            return self.dispatch_line(frame)
        if event == "call":
            return self.dispatch_call(frame, arg)
        # if event == 'return':
        #     return self.dispatch_return(frame, arg)
        # if event == 'exception':
        #     return self.dispatch_exception(frame, arg)
        # print('bdb.Bdb.dispatch: unknown debugging event:', repr(event))
        return self.trace_dispatch

    def dispatch_line(self, frame):
        if self.stop_here(frame) or self.break_here(frame):
            self.user_line(frame)
            if self.quitting:
                raise MpdbQuit
        return self.trace_dispatch

    def dispatch_call(self, frame, arg):
        if self.botframe is None:
            # First call of dispatch since reset()
            self.botframe = frame.f_back  # (CT) Note that this may also be None!
            return self.trace_dispatch
        if not (self.stop_here(frame) or self.break_anywhere(frame)):
            # No need to trace this function
            return  # None
        self.user_call(frame, arg)
        return self.trace_dispatch

    # def dispatch_return(self, frame, arg):
    #     pass

    def get_break(self, filename, lineno):
        """Return True if there is a breakpoint for filename:lineno."""
        return filename in self.breaks and lineno in self.breaks[filename]

    def get_breaks(self, filename, lineno):
        """Return all breakpoints for filename:lineno.
        If no breakpoints are set, return an empty list.
        """
        return (
            (filename in self.breaks
            and lineno in self.breaks[filename]
            and Breakpoint.bplist[filename, lineno])
            or []
        )

    def break_here(self, frame):
        filename = frame.f_code.co_filename
        if filename not in self.breaks:
            return False
        lineno = frame.f_lineno
        if lineno not in self.breaks[filename]:
            # The line itself has no breakpoint, but maybe the line is the
            # first line of a function with breakpoint set by function name.
            lineno = frame.f_code.co_firstlineno
            if lineno not in self.breaks[filename]:
                return False

        # flag says ok to delete temp. bp
        bp, flag = effective(filename, lineno, frame)
        if bp:
            self.currentbp = bp.number
            if flag and bp.temporary:
                self.do_clear(str(bp.number))
            return True
        else:
            return False

    def do_clear(self, arg):
        raise NotImplementedError("Subclass of BDB must implement this functionality")

    def clear_all_breaks(self):
        """Delete all existing breakpoints.
        If none were set, return an error message.
        """
        if not self.breaks:
            return "There are no breakpoints"
        for bp in Breakpoint.bpbynumber:
            if bp:
                bp.deleteMe()
        self.breaks = {}
        return None

    def stop_here(self, frame):
        if frame is self.stopframe:
            if self.stoplineno == -1:
                return False
            return frame.f_lineno >= self.stoplineno
        if not self.stopframe:
            return True
        return False

    def set_next(self, frame):
        self._set_stopinfo(frame, None)

    def set_step(self):
        self._set_stopinfo(None, None)

    def set_return(self, frame):
        self._set_stopinfo(frame.f_back, frame)

    def set_continue(self):
        """Stop only at breakpoints or when finished.
        If there are no breakpoints, set the system trace function to None.
        """
        # Don't stop except at breakpoints or when finished
        self._set_stopinfo(self.botframe, None, -1)
        if not self.breaks:
            sys.settrace(None)

    def set_until(self, frame, lineno=None):
        """Stop when the line with the lineno greater than the current one is
        reached or when returning from current frame."""
        # the name "until" is borrowed from gdb
        if lineno is None:
            lineno = frame.f_lineno + 1
        self._set_stopinfo(frame, frame, lineno)

    def set_break(self, filename, lineno, temporary=False, cond=None):
        bp_list = self.breaks.setdefault(filename, [])
        if lineno not in bp_list:
            bp_list.append(lineno)
        Breakpoint(filename, lineno, temporary, cond)
        return None

    def set_quit(self):
        """Set quitting attribute to True.
        Raises BdbQuit exception in the next call to a dispatch_*() method.
        """
        self.stopframe = self.botframe
        self.returnframe = None
        self.quitting = True
        sys.settrace(None)

    def break_anywhere(self, frame) -> bool:
        return frame.f_code.co_filename in self.breaks

    def clear_break(self, filename, lineno):
        """Delete breakpoints for filename:lineno.
        If no breakpoints were set, return an error message.
        """
        if filename not in self.breaks:
            return f"There are no breakpoints in {filename}"
        if lineno not in self.breaks[filename]:
            return f"There is no breakpoint at {filename}:{lineno}"
        for bp in Breakpoint.bplist[filename, lineno][:]:
            bp.deleteMe()
        self._prune_breaks(filename, lineno)

    def clear_bpbynumber(self, arg):
        """Delete a breakpoint by its index in Breakpoint.bpbynumber.
        If arg is invalid, return an error message.
        """
        try:
            bp = self.get_bpbynumber(arg)
        except ValueError as err:
            return str(err)
        bp.deleteMe()
        self._prune_breaks(bp.file, bp.line)
        return None

    def get_bpbynumber(self, arg):
        """Return a breakpoint by its index in Breakpoint.bybpnumber.
        For invalid arg values or if the breakpoint doesn't exist,
        raise a ValueError.
        """
        if not arg:
            raise ValueError("Breakpoint number expected")
        try:
            number = int(arg)
        except ValueError:
            raise ValueError(f"Non-numeric breakpoint number {arg}") from None
        try:
            bp = Breakpoint.bpbynumber[number]
        except IndexError:
            raise ValueError(f"Breakpoint number {number} out of range") from None
        if bp is None:
            raise ValueError(f"Breakpoint {number} already deleted")
        return bp

    def _prune_breaks(self, filename, lineno):
        """Prune breakpoints for filename:lineno.
        A list of breakpoints is maintained in the Bdb instance and in
        the Breakpoint class.  If a breakpoint in the Bdb instance no
        longer exists in the Breakpoint class, then it's removed from the
        Bdb instance.
        """
        if (filename, lineno) not in Breakpoint.bplist:
            self.breaks[filename].remove(lineno)
        if not self.breaks[filename]:
            del self.breaks[filename]

    def _set_stopinfo(self, stopframe, returnframe, stoplineno=0):
        """Set the attributes for stopping.
        If stoplineno is greater than or equal to 0, then stop at line
        greater than or equal to the stopline.  If stoplineno is -1, then
        don't stop at all.
        """
        self.stopframe = stopframe
        self.returnframe = returnframe
        self.quitting = False
        # stoplineno >= 0 means: stop at line >= the stoplineno
        # stoplineno -1 means: don't stop at all
        self.stoplineno = stoplineno

    def set_trace(self):
        sys.settrace(self.trace_dispatch)

    def user_line(self, frame):
        pass

    def user_call(self, frame, arg):
        pass


# Determines if there is an effective (active) breakpoint at this
# line of code.  Returns breakpoint number or 0 if none
def effective(file, line, frame):
    """Determine which breakpoint for this file:line is to be acted upon.
    Called only if we know there is a breakpoint at this location.  Return
    the breakpoint that was triggered and a boolean that indicates if it is
    ok to delete a temporary breakpoint.  Return (None, None) if there is no
    matching breakpoint.
    """
    possibles = Breakpoint.bplist[file, line]
    for b in possibles:
        if not b.enabled:
            continue
        # Count every hit when bp is enabled
        b.hits += 1
        if not b.cond:
            # If unconditional, and ignoring go on to next, else break
            if b.ignore > 0:
                b.ignore -= 1
                continue
            else:
                # breakpoint and marker that it's ok to delete if temporary
                return b, True
        else:
            # Conditional bp.
            # Ignore count applies only to those bpt hits where the
            # condition evaluates to true.
            try:
                # Even tho MicroPython eval statement is limited
                # to just globals it is till useful
                val = eval(b.cond, frame.f_globals, frame.f_globals)
                if val:
                    if b.ignore > 0:
                        b.ignore -= 1
                        # continue
                    else:
                        return b, True
                # else:
                #   continue
            except Exception:
                # if eval fails, most conservative thing is to stop on
                # breakpoint regardless of ignore count.  Don't delete
                # temporary, as another hint to user.
                return b, False
    return None, None


# noinspection DuplicatedCode
class Breakpoint:
    """Breakpoint class.
    Implements temporary breakpoints, ignore counts, disabling and
    (re)-enabling, and conditionals.
    Breakpoints are indexed by number through bpbynumber and by
    the (file, line) tuple using bplist.  The former points to a
    single instance of class Breakpoint.  The latter points to a
    list of such instances since there may be more than one
    breakpoint per line.
        A conditional breakpoint always counts a hit.
    """

    # XXX Keeping state in the class is a mistake -- this means
    # you cannot have more than one active Bdb instance.

    next = 1  # Next bp to be assigned
    bplist: ClassVar[dict] = {}  # indexed by (file, lineno) tuple
    bpbynumber: ClassVar[list] = [None]  # Each entry is None or an instance of Bpt

    # index 0 is unused, except for marking an
    # effective break .... see effective()

    def __init__(self, file, line, temporary=False, cond=None):
        self.func_first_executable_line = None
        self.file = file  # This better be in canonical form!
        self.line = line
        self.temporary = temporary
        self.cond = cond
        self.enabled = True
        self.ignore = 0
        self.hits = 0
        self.number = Breakpoint.next
        Breakpoint.next += 1
        # Build the two lists
        self.bpbynumber.append(self)
        if (file, line) in self.bplist:
            self.bplist[file, line].append(self)
        else:
            self.bplist[file, line] = [self]

    def deleteMe(self):
        """Delete the breakpoint from the list associated to a file:line.
        If it is the last breakpoint in that position, it also deletes
        the entry for the file:line.
        """

        index = (self.file, self.line)
        self.bpbynumber[self.number] = None  # No longer in list
        self.bplist[index].remove(self)
        if not self.bplist[index]:
            # No more bp for this f:l combo
            del self.bplist[index]

    def enable(self):
        """Mark the breakpoint as enabled."""
        self.enabled = True

    def disable(self):
        """Mark the breakpoint as disabled."""
        self.enabled = False

    def bpprint(self, out=None):
        """Print the output of bpformat().
        The optional out argument directs where the output is sent
        and defaults to standard output.
        """
        if out is None:
            out = sys.stdout
        print(self.bpformat(), file=out)

    def bpformat(self):
        """Return a string with information about the breakpoint.
        The information includes the breakpoint number, temporary
        status, file:line position, break condition, number of times to
        ignore, and number of times hit.
        """
        disp = "del  " if self.temporary else "keep "
        disp += "yes  " if self.enabled else "no   "
        ret = f"{self.number:-4d}breakpoint   {disp} at {self.file}:{self.line}"
        if self.cond:
            ret += f"\n\tstop only if {self.cond}"
        if self.ignore:
            ret += f"\n\tignore next {self.ignore} hits"
        if self.hits:
            ss = "s" if self.hits > 1 else ""
            ret += f"\n\tbreakpoint already hit {self.hits} time{ss}"
        return ret

    def __str__(self):
        """Return a condensed description of the breakpoint."""
        return f"breakpoint {self.number} at {self.file}:{self.line}"


# Adapted from: https://github.com/micropython/micropython-lib/blob/master/cmd/cmd.py
class Cmd:
    IDENTCHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"
    PROMPT = "(Mpdb) "
    doc_leader = ""
    doc_header = "Documented commands (type help <topic>):"
    misc_header = "Miscellaneous help topics:"
    undoc_header = "Undocumented commands:"
    nohelp = "*** No help on %s"
    ruler = "="

    def __init__(self):
        self.lastcmd = ""

    def cmdloop(self):
        try:
            stop = None
            while not stop:
                try:
                    line = input(self.PROMPT)
                except EOFError:
                    line = "EOF"
                stop = self.onecmd(line)
        finally:
            pass

    def onecmd(self, line):
        """Interpret the argument as though it had been typed in response
        to the prompt.
        """
        cmd, arg, line = self.parseline(line)
        if not line:
            return self.emptyline()
        if cmd is None:
            return self.default(line)
        self.lastcmd = line
        if line == "EOF":
            self.lastcmd = ""
        if cmd == "":
            return self.default(line)
        else:
            try:
                func = getattr(self, "do_" + cmd)
            except AttributeError:
                return self.default(line)
            return func(arg)

    def parseline(self, line):
        """Parse the line into a command name and a string containing
        the arguments.  Returns a tuple containing (command, args, line).
        'command' and 'args' may be None if the line couldn't be parsed.
        """
        line = line.strip()
        if not line:
            return None, None, line
        elif line[0] == "?":
            line = "help " + line[1:]
        elif line[0] == "!":
            if hasattr(self, "do_shell"):
                line = "shell " + line[1:]
            else:
                return None, None, line
        i, n = 0, len(line)
        while i < n and line[i] in self.IDENTCHARS:
            i = i + 1
        cmd, arg = line[:i], line[i:].strip()
        return cmd, arg, line

    def default(self, line):
        """Called on an input line when the command prefix is not recognized.
        If this method is not overridden, it prints an error message and
        returns.
        """
        sys.stdout.write(f"*** Unknown syntax: {line}\n")

    def emptyline(self):
        """Called when an empty line is entered in response to the prompt.
        If this method is not overridden, it repeats the last nonempty
        command entered.
        """
        if self.lastcmd:
            return self.onecmd(self.lastcmd)

    def get_names(self):
        # This method used to pull in base class attributes
        # at a time dir() didn't do it yet.
        return dir(self.__class__)

    def do_help(self, arg):
        """List available commands with "help" or detailed help with "help cmd"."""
        if arg:
            # XXX check arg syntax
            try:
                func = getattr(self, "help_" + arg)
            except AttributeError:
                sys.stdout.write(f"{self.nohelp % (arg,)!s}\n")
                return
            func()
        else:
            names = self.get_names()
            cmds_doc = []
            cmds_undoc = []
            help = {}
            for name in names:
                if name[:5] == "help_":
                    help[name[5:]] = 1
            names.sort()
            # There can be duplicates if routines overridden
            prevname = ""
            for name in names:
                if name[:3] == "do_":
                    if name == prevname:
                        continue
                    prevname = name
                    cmd = name[3:]
                    if cmd in help:
                        cmds_doc.append(cmd)
                        del help[cmd]
                    else:
                        cmds_undoc.append(cmd)
            sys.stdout.write(f"{self.doc_leader!s}\n")
            self.print_topics(self.doc_header, cmds_doc, 15, 80)
            self.print_topics(self.misc_header, list(help.keys()), 15, 80)
            self.print_topics(self.undoc_header, cmds_undoc, 15, 80)

    def print_topics(self, header, cmds, cmdlen, maxcol):  # noqa:ARG002
        if cmds:
            sys.stdout.write(f"{header!s}\n")
            if self.ruler:
                sys.stdout.write(f"{self.ruler * len(header)!s}\n")
            for cmd in cmds:
                print(cmd)
            sys.stdout.write("\n")


def print_stacktrace(frame, level=0):
    print(
        "%2d: %s@%s:%s => %s:%d"
        % (
            level,
            "  ",
            frame.f_globals["__name__"],
            frame.f_code.co_name,
            frame.f_code.co_filename,
            frame.f_lineno,
        ),
    )

    if frame.f_back:
        print_stacktrace(frame.f_back, level + 1)


class Mpdb(Mbdb, Cmd):
    def do_break(self, arg, temporary=False):
        if not arg:
            if self.breaks:  # There's at least one
                self.message("Num Type         Disp Enb   Where")
                for bp in Breakpoint.bpbynumber:
                    if bp:
                        self.message(bp.bpformat())
            return
        filename = None
        lineno = None
        cond = None
        comma = arg.find(",")
        if comma > 0:
            # parse stuff after comma: "condition"
            cond = arg[comma + 1 :].lstrip()
            arg = arg[:comma].rstrip()
        # parse stuff before comma: [filename:]lineno | function
        colon = arg.rfind(":")
        if colon >= 0:
            filename = arg[:colon].rstrip()
            print(filename)
            arg = arg[colon + 1 :].lstrip()
            try:
                lineno = int(arg)
            except ValueError:
                return
        self.set_break(filename, lineno, cond=cond, temporary=temporary)

    def do_clear(self, arg):
        """cl(ear) filename:lineno\ncl(ear) [bpnumber [bpnumber...]]
        With a space separated list of breakpoint numbers, clear
        those breakpoints.  Without argument, clear all breaks (but
        first ask confirmation).  With a filename:lineno argument,
        clear all breaks at that line in that file.
        """
        if not arg:
            try:
                reply = input("Clear all breaks? ")
            except EOFError:
                reply = "no"
            reply = reply.strip().lower()
            if reply in ("y", "yes"):
                bplist = [bp for bp in Breakpoint.bpbynumber if bp]
                self.clear_all_breaks()
                for bp in bplist:
                    self.message(f"Deleted {bp}")
            return
        if ":" in arg:
            # Make sure it works for "clear C:\foo\bar.py:12"
            i = arg.rfind(":")
            filename = arg[:i]
            arg = arg[i + 1 :]
            try:
                lineno = int(arg)
            except ValueError:
                err = f"Invalid line number ({arg})"
            else:
                bplist = self.get_breaks(filename, lineno)
                err = self.clear_break(filename, lineno)
            if err:
                self.error(err)
            else:
                for bp in bplist:
                    self.message(f"Deleted {bp}")
            return
        numberlist = arg.split()
        for i in numberlist:
            try:
                bp = self.get_bpbynumber(i)
            except ValueError as err:
                self.error(err)
            else:
                self.clear_bpbynumber(i)
                self.message(f"Deleted {bp}")

    do_cl = do_clear  # 'c' is already an abbreviation for 'continue'

    def do_step(self, arg):  # noqa:ARG002
        """s(tep)
        Execute the current line, stop at the first possible occasion
        (either in a function that is called or in the current
        function).
        """
        self.set_step()
        return 1

    do_s = do_step

    def do_next(self, arg):  # noqa:ARG002
        """n(ext)
        Continue execution until the next line in the current function
        is reached or it returns.
        """
        self.set_next(self.curframe)
        return 1

    do_n = do_next

    def do_continue(self, arg):  # noqa:ARG002
        """c(ont(inue))
        Continue execution, only stop when a breakpoint is encountered.
        """
        self.set_continue()
        return 1

    do_c = do_cont = do_continue

    def do_where(self, arg):  # noqa:ARG002
        """w(here)
        Print a stack trace.
        """
        print_stacktrace(self.curframe)
        self.cmdloop()
        return 1

    do_w = do_where

    def do_until(self, arg):
        """unt(il) [lineno]
        Without argument, continue execution until the line with a
        number greater than the current one is reached.  With a line
        number, continue execution until a line with a number greater
        or equal to that is reached.
        """
        if arg:
            try:
                lineno = int(arg)
            except ValueError:
                self.error(f"Error in argument: {arg!r}")
                return
            if lineno <= self.curframe.f_lineno:
                self.error('"until" line number is smaller than current line number')
                return
        else:
            lineno = None
        self.set_until(self.curframe, lineno)
        return 1

    do_unt = do_until

    def do_tbreak(self, arg):
        self.do_break(arg, True)

    def message(self, msg):
        print(msg)

    def error(self, msg):
        print("***", msg)

    def do_quit(self, arg):  # noqa:ARG002
        self.set_quit()
        return 1

    do_q = do_quit

    def do_mem_free(self, arg):  # noqa:ARG002
        print("Memory Free: ", gc.mem_free())

    do_mf = do_mem_free

    def do_mem_alloc(self, arg):  # noqa:ARG002
        print("Memory Allocated: ", gc.mem_alloc())

    do_ma = do_mem_alloc

    def do_collect(self, arg):  # noqa:ARG002
        gc.collect()
        print(
            "Garbage Collected!",
        )

    def do_p(self, arg):
        repr(self._getval(arg))

    def _getval(self, arg):
        return eval(arg, self.curframe.f_globals, self.curframe.f_globals)

    def user_line(self, frame):
        out_info = f"{frame.f_code.co_filename}:{frame.f_lineno}"
        if self.currentbp != 0:
            out_info = "BREAK " + out_info
            self.currentbp = 0
        else:
            out_info = "STOP " + out_info
        print(out_info)
        self.cmdloop()


_db = None


def breakpoint():  # noqa:A001
    "mimic CPython's `breakpoint`"
    global _db
    if _db is None:
        _db = Mpdb()
    _db.set_trace()



"""
Helpers for MoaT command interpreters et al.
"""

from __future__ import annotations

from moat.util import NotGiven

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any


def get_part(cur: Any, p: Sequence[str | int], add: bool = False) -> Any:
    """Walk into a mapping or object structure."""
    for pp in p:
        try:
            cur = getattr(cur, pp)  # type: ignore[arg-type]  # pp might be int but then getattr raises
        except (TypeError, AttributeError):
            try:
                cur = cur[pp]  # type: ignore[invalid-argument-type]
            except TypeError:
                if pp is None:
                    raise KeyError(p, pp) from None
                raise
            except KeyError:
                if not add:
                    raise KeyError(p, pp) from None
                cur[pp] = nc = []  # type: ignore[invalid-assignment]
                cur = nc
    return cur


def set_part(cur: Any, p: Sequence[str | int], v: Any, keep: bool = False) -> None:
    """Modify a mapping or object structure."""
    cur = get_part(cur, p[:-1], add=True)
    pp = p[-1]

    if v is NotGiven and not keep:
        try:
            cur.pop(pp, None)
        except (TypeError, AttributeError):
            if pp is None or (isinstance(pp, int) and len(cur) == pp):
                cur.pop()
            else:
                cur.remove(pp)

    else:
        try:
            cur[pp] = v
        except (TypeError, AttributeError):
            if pp is None or (isinstance(pp, int) and len(cur) == pp):
                cur.append(v)
            else:
                setattr(cur, pp, v)  # type: ignore[arg-type]  # pp might be int but then we append above


def enc_part(
    cur: Any, name: str | None = None
) -> tuple[Any, tuple[Any, ...] | None] | tuple[Any, Sequence[int | str]] | Any:
    """
    Helper method to encode a larger dict/list partially.

    The result is either some object that's not a dict or list, or a
    (X,L) tuple where X is the dict/list in question except with all the
    complex parts removed, and L is a list of keys/offsets with complex
    data to retrieve

    The tuple may have a third element: the name, if passed in.
    """

    def _complex(v: Any) -> bool:
        if isinstance(v, (dict, list, tuple)):
            return True
        return False

    if isinstance(cur, dict):
        c: dict[Any, Any] = {}
        s: list[str | int] = []
        for k, v in cur.items():
            if _complex(v):
                s.append(k)
            else:
                c[k] = v
        if s or name:
            return (c, s, name) if name else (c, s)
        else:
            # dict has no complex values: return directly
            return c

    elif isinstance(cur, (list, tuple)):
        c_list: list[Any] = []
        s_list: list[int] = []
        for k, v in enumerate(cur):
            if _complex(v):
                c_list.append(None)
                s_list.append(k)
            else:
                c_list.append(v)
        # cannot do a direct return here
        return c_list, s_list

    else:
        return cur
        # guaranteed not to be a tuple



"""
HDC2010 i²c
"""

from __future__ import annotations

import struct

from moat.lib.micro import Lock, sleep_ms
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import Msg

__all__ = ["Cmd"]


class Cmd(BaseCmd):
    """
    This is the front-end for a HDC2010 i²c temperature+humidity sensor.

    Parameters:
        bus: Path to the bus to use
        adr: Bus address to access
    """

    lock: Lock

    doc = dict(_c=dict(_d="HDC2010 temp+humid sensor", bus="path:i2c bus", adr="int:bus address"))

    async def setup(self):
        "allocate lock"
        super().setup()
        self.lock = Lock()

    async def task(self):
        "wrapper, for i2c bus access"
        self.adr = self.cfg["adr"]
        async with self.root.sub_at(self.cfg["bus"]) as self.bus:
            await super().task()

    doc_r = dict(
        _d="read",
        t="int:ms between reads when streaming, default 10s",
        o="bool:old: wait until value differs",
    )

    async def _rdt(self):
        async with self.lock:
            await self.bus.wr(self.adr, bytes((15, 0x01)))
            # OneShot+Shutdown, Extended
            for _ in range(20):
                await sleep_ms(3)
                res = await self.bus.wrrd(self.adr, bytes((4,)), 1)
                if res[0] & 0x80:
                    break
            else:
                raise RuntimeError("No conv")
            res = await self.bus.wrrd(self.adr, bytes((0,)), 2)
        (t,) = struct.unpack("<H", res)
        return t

    async def stream_rt(self, msg: Msg):
        "Wait for change if @o (old value) is not None"
        t = msg.get("t", 10000)
        o = msg.get("o", None)
        if o is not None:
            o = int((o + 40) * 65536 / 165)
        d = int(msg.get("d", 0) * 65536 / 165)
        if msg.can_stream:
            async with msg.stream_out() as m:
                while True:
                    val = await self._rdt()
                    if o is None or abs(val - o) > d:
                        await m.send(val * 165 / 65536 - 40)
                        if o is not None:
                            o = val
                    await sleep_ms(t)

        val = await self._rdt()
        while o is not None and abs(val - o) <= d:
            await sleep_ms(t)
            val = await self._rdt()
        await msg.result(val * 165 / 65536 - 40)

    async def _rdh(self):
        async with self.lock:
            await self.bus.wr(self.adr, bytes((15, 0x01)))
            # OneShot+Shutdown, Extended
            for _ in range(20):
                await sleep_ms(3)
                res = await self.bus.wrrd(self.adr, bytes((4,)), 1)
                if res[0] & 0x80:
                    break
            else:
                raise RuntimeError("No conv")
            res = await self.bus.wrrd(self.adr, bytes((2,)), 2)
        (h,) = struct.unpack("<H", res)
        return h

    async def stream_rh(self, msg: Msg):
        "Wait for change if @o (old value) is not None"
        t = msg.get("t", 10000)
        o = msg.get("o", None)
        if o is not None:
            o = int(o * 65536 / 100)
        d = int(msg.get("d", 0) * 65536 / 100)
        if msg.can_stream:
            async with msg.stream_out() as m:
                while True:
                    val = await self._rdh()
                    if o is None or abs(val - o) > d:
                        await m.send(val * 100 / 65536)
                        if o is not None:
                            o = val
                    await sleep_ms(t)

        val = await self._rdh()
        while o is not None and abs(val - o) <= d:
            await sleep_ms(t)
            val = await self._rdh()
        await msg.result(val * 100 / 65536)

    async def _rdth(self):
        async with self.lock:
            await self.bus.wr(self.adr, bytes((15, 0x01)))
            # OneShot+Shutdown, Extended
            for _ in range(20):
                await sleep_ms(3)
                res = await self.bus.wrrd(self.adr, bytes((4,)), 1)
                if res[0] & 0x80:
                    break
            else:
                raise RuntimeError("No conv")
            res = await self.bus.wrrd(self.adr, bytes((0,)), 4)
        (t, h) = struct.unpack("<HH", res)
        return t, h

    async def stream_rth(self, msg: Msg):
        "Wait for change if @o (old value) is not None"
        t = msg.get("t", 10000)
        ot = msg.get("ot", None)
        if ot is not None:
            ot = int((ot + 40) * 65536 / 165)
        oh = msg.get("oh", None)
        if oh is not None:
            oh = int(oh * 65536 / 100)
        dt = int(msg.get("d", 0) * 65536 / 165)
        dh = int(msg.get("d", 0) * 65536 / 100)
        if msg.can_stream:
            async with msg.stream_out() as m:
                while True:
                    t, h = await self._rdth()
                    if ot is None or oh is None or abs(t - ot) > dt or abs(h - oh) > dh:
                        await m.send(t * 165 / 65536 - 40, h * 100 / 65536)
                        if ot is not None:
                            ot = t
                        if oh is not None:
                            oh = h
                    await sleep_ms(t)

        t, h = await self._rdth()
        while ot is not None and oh is not None and abs(t - ot) <= dt and abs(h - oh) <= dh:
            await sleep_ms(t)
            t, h = await self._rdth()
        await msg.result(t * 165 / 65536 - 40, h * 100 / 65536)



"""
Access a satellite's i²c bus.
"""

from __future__ import annotations

from smbus3 import SMBus, i2c_msg

from moat.lib.micro import Lock, to_thread
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from moat.lib.path import PathElem
    from moat.lib.rpc import Msg

    from typing import Any, Protocol

    class _I2CWriteRes(Protocol):
        len: int


class Cmd(BaseCmd):
    r"""
    This command implements basic access to an I²C bus.

    Parameters::

        id: bus ID (no soft i2c here)
        \# f: 100000  \# frequency, Hz
        t: 1000 \# bus timeout, msec
    """

    doc = dict(
        c=dict(
            _d="i²c bus read/write/dir",
            t="int:timeout(ms)",
            id="int:bus#",
        )
    )

    _bus: SMBus

    def __init__(self, cfg: dict):
        super().__init__(cfg)
        self._lock = Lock()

    async def setup(self):
        """
        Open a bus.
        """
        await super().setup()
        self._setup()

    async def reload(self):
        "reconfigured"
        self._teardown()
        self._setup()
        await super().reload()

    def _setup(self):
        cfg = self.cfg
        # f = cfg.get("f", 100000)
        t = cfg.get("t", 1000) * 1000
        self._bus = b = SMBus(cfg.id)
        b.set_timeout(int(t * 100))

    async def teardown(self):
        "shutdown"
        self._teardown()
        await super().teardown()

    def _teardown(self):
        try:
            b = self._bus
        except AttributeError:
            return
        del self._bus
        b.close()

    doc_rd = dict(_d="read", _0="int:addr", n="int:nbytes(16)")

    async def cmd_rd(self, i, n=16) -> bytes:
        "read @n bytes from bus @cd at address @i"
        return bytes(self._bus.i2c_rd(i, n))

    doc_wr = dict(_d="write", _0="int:addr", buf="bytes:data", _r="int:nbytes")

    async def cmd_wr(self, i: int, buf: bytes) -> int:
        "write @buf to bus @cd at address @i"
        async with self._lock:
            res = cast("_I2CWriteRes", await to_thread(self._bus.i2c_wr, i, list(buf)))
            return res.len

    doc_wrrd = dict(
        _d="write+read",
        _0="int:addr",
        buf="bytes:data",
        n="int:nbytes(16)",
        _r="int|bytes:nbytes short-written|read result",
    )

    async def cmd_wrrd(self, i, buf, n=16) -> bytes:
        """
        write @buf to bus @cd at address @i, then read @n bytes.

        Returns -x if only x bytes could be written.
        """

        def _run():
            wr = i2c_msg.write(i, list(buf))
            rd = i2c_msg.read(i, n)

            self._bus.i2c_rdwr(wr, rd)
            return rd

        async with self._lock:
            rd = await to_thread(_run)
        return bytes(rd)

    doc_scan = dict(_d="bus scan")

    async def cmd_scan(self) -> list[int]:
        """
        Scan the bus.

        Returns: the list of working addresses.
        """
        bus = self._bus
        res = []
        for i in range(0x08, 0x78):
            try:
                if i >> 4 in (3, 5):
                    await to_thread(bus.read_byte, i)
                else:
                    await to_thread(bus.write_quick, i)
            except OSError:
                pass
            else:
                res.append(i)
        return res

    async def handle(self, msg: Msg, rcmd: list[PathElem]) -> Any:
        """
        Intercept the bus address: ``bus:33.wr(x)`` ≍ ``bus.wr(33,x)``.
        """
        if rcmd and isinstance(rcmd[-1], int):
            msg.args_l.insert(0, rcmd.pop())
        return await super().handle(msg, rcmd)



#
# Created on Wed Jun 22 20:06:38 2022
#
# @author: eadali
#
# extended for MoaT by Matthias Urlichs
#

from __future__ import annotations

from math import exp

from moat.util import NotGiven, attrdict
from moat.lib.micro import ticks_diff

from typing import TYPE_CHECKING

__all__ = []

if TYPE_CHECKING:
    from types import EllipsisType

    from collections.abc import Sequence

try:
    # MicroPython. Use milliseconds. Type resolved below.
    from time import ticks_ms as time  # type:ignore[unresolved-import]

    PID_TC = 1000
    MAX_VAL = 2**31 - 1

except ImportError:
    # Standard Python
    from time import monotonic as time

    PID_TC = 1
    MAX_VAL = float("inf")


class PID:
    """
    An advanced PID controller with first-order filter on derivative term.
    """

    t: float | None = None
    e: float | None = None
    _err: float | None = None
    i: float = 0

    Kp: float
    Ki: float | None
    Kd: float | None
    Tf: float | None

    lower: float = -MAX_VAL
    upper: float = MAX_VAL

    fct: float = 1
    off: float = 0

    def __init__(
        self,
        Kp: float | None = None,
        Ki: float | None = None,
        Kd: float | None = None,
        Tf: float | None = None,
        t: float | None = None,
    ):
        """
        Args:
            Kp: proportional term
            Ki: integral term.
            Kd: derivative term.
            Tf: Time constant for first-order filter on the derivative
            t: current / initial time

        Attributes:
            t: current time
            e: current differential error
            i: current integral sum

            lower: Lower limit of the output
            upper: Upper limit of the output
        """
        self.set_gains(Kp, Ki, Kd, Tf)
        self.set_output_limits()
        self.reset(t)

    def reset(self, t: float | None = None, clear: bool = False) -> None:
        """
        Reset the controller.

        Args:
            t:
                current time (if known)
            clear:
                if set, clear internal state so that ``self(0) == 0``
        """
        self.set_state(t, 0 if clear else None, 0 if clear else None)
        if t is None:
            self.t = None

    def __call__(self, e: float, t: float | None = None) -> float:
        """Run a PID step.

        Args:
            e: Error signal.
            t: Current time.

        Returns:
            Control signal.

        """
        return self.sum(self.integrate(e, t))

    def sum(self, args: Sequence[float]) -> float:
        """
        Sum the input values, apply fct+off, and limit the result to the
        interval between `min`…`max`.

        ``pid(signal)`` ≍ ``pid.sum(*self.integrate(signal))``
        """
        return self.off + self.fct * min(max(sum(args), self.lower), self.upper)

    def get_gains(self) -> tuple[float, float | None, float | None, float | None]:
        """Get PID controller gains.

        Returns
        -------
        tuple
            Gains of PID controller (Kp, Ki, Kd, Tf).

        """
        return (
            self.Kp,
            self.Ki * PID_TC if self.Ki else None,
            self.Kd / PID_TC if self.Kd else None,
            self.Tf / PID_TC if self.Tf else None,
        )

    def set_offset(self, fct: float | None = None, off: float | None = None) -> None:
        """Set output value factor and offset.

        Args:
            fct: Factor (default 1.0)
            off: Offset (default 0.0)

        Factor and offset are applied by :meth:`sum`.
        """
        if fct is not None:
            self.fct = fct
        if off is not None:
            self.off = off

    def get_offset(self) -> tuple[float, float]:
        """Get output value factor and offset."""
        return self.fct, self.off

    def set_gains(
        self, Kp: float | None, Ki: float | None, Kd: float | None, Tf: float | None = None
    ):
        """Set PID controller gains.

        Args:
            Kp: Proportional gain.
            Ki: Integral gain.
            Kd: Derivative gain.
            Tf: Time constant of the first-order derivative filter.

        """
        old_Kp = getattr(self, "Kp", 0.0)
        old_Kd = getattr(self, "Kd", None)
        old_Tf = getattr(self, "Tf", None)

        new_Kp = Kp or 0
        new_Ki = Ki / PID_TC if Ki else None
        new_Kd = Kd * PID_TC if Kd else None
        new_Tf = Tf * PID_TC if Tf else None

        err = self._err
        if err is None:
            err = self.e

        if err is not None:
            # keep p+i continuous across gain changes
            self.i += (old_Kp - new_Kp) * err

            if old_Kd and old_Tf and self.e is not None:
                if new_Kd and new_Tf:
                    # Adjust derivative filter residual for a bumpless D/Tf change.
                    tdf0 = old_Tf / old_Kd
                    tdf1 = new_Tf / new_Kd
                    self.e = err - (tdf1 / tdf0) * (err - self.e)
                else:
                    self.e = err
            elif new_Kd and new_Tf:
                self.e = err

        self.Kp = new_Kp
        self.Ki = new_Ki
        self.Kd = new_Kd
        self.Tf = new_Tf

    def set_output_limits(self, lower: float | None = None, upper: float | None = None):
        """Set PID controller output limits, for anti-windup.

        Args:
            lower: Lower limit for anti-windup,
            upper: Upper limit for anti-windup.

        """
        if lower is not None:
            self.lower = lower
        if upper is not None:
            self.upper = upper

    def get_output_limits(self) -> tuple[float, float]:
        """Get PID controller output limits for anti-windup.

        Returns: Output limits (lower, upper).

        """
        return self.lower, self.upper

    def set_state(
        self,
        t: float | None | EllipsisType = None,
        e: float | None = None,
        i: float | None = None,
    ) -> None:
        """Set PID controller states.

        Args:
            t:
                Current time. If NotGiven, clears.
            e:
                Current error.
            i:
                Current integral.

        """
        if t is NotGiven:
            self.t = None
        elif t is not None:
            self.t = t
        if e is not None:
            self.e = e
            self._err = e
        if i is not None:
            self.i = i

    def get_state(self) -> tuple[float | None, float | None, float]:
        """Get PID controller states.

        Returns:
            State of the PID controller (t, e, i)

        """
        return self.t, self.e, self.i

    def integrate(self, e: float, t: float | None = None) -> tuple[float, float, float]:
        """Calculates PID controller output.

        Args:
            e: Error signal.
            t: Current time.

        Returns:
            ``p,i,d``: Control signal (in parts), *not* limited.

        This method performs anti-windup protection on the controller's integral term.
        """
        e1 = e
        t0, e0, i0 = self.get_state()
        if t is None:
            t = time()

        if t0 is None:
            t0 = t
        if e0 is None:
            e0 = e

        # Calculate time step
        dt = ticks_diff(t, t0)
        if dt < 0:
            self.t = t
            raise ValueError(f"Time went backwards: {t} {t0} {dt}")

        # Calculate proportional term
        p = self.Kp * e
        # Calculate integral term
        if self.Ki:
            i = i0 + dt * self.Ki * e
            # anti-windup
            i = min(max(i, min(self.lower - p, 0)), max(self.upper - p, 0))
        else:
            i = i0

        # Calculate possibly-delayed derivative term
        d = 0.0
        if dt and (Kd := self.Kd):
            if (Tf := self.Tf) is not None:
                Tdf = Tf / Kd
                x = -e0 / Tdf
                ex = exp(-dt / Tf)
                y = ex * x - (1.0 - ex) / Tdf * e
                d = y + e / Tdf
                e = -Tdf * y
            else:
                d = (e - e0) * Kd / dt
        # Set initial value for next cycle
        self.set_state(t, e, i)
        self._err = e1

        return p, i, d


class CPID(PID):
    """
    A PID that's configured with a config dictionary.

    This class accepts the desired and actual system state as input, *not* the error.
    ::

        flow:
            p: 0.1
            i: 0.01
            d: 0.0
            tf: 0.0  # both d ant tf must be set

            # output limits
            min: .3
            max: .95

            # setpoint change: adjust integral for best guess
            # input 20, output .8 == 20/.8
            factor: .04
            offset: 0

            state: foo
    """

    def __init__(self, cfg: attrdict, state: attrdict | None = None, t: float | None = None):
        """
        Args:
            cfg: our configuration. See above.
            state: the state storage. Typically an :py:class:`moat.util.attrdict`.
        """
        super().__init__(cfg.get("p", 0), cfg.get("i", 0), cfg.get("d", 0), cfg.get("tf", 0))
        self.cfg = cfg
        self.cfg_updated()

        if state is None:
            state = attrdict()
        self.state = state
        self.set_state(state.get("t", t or time()), state.get("e", 0), state.get("i", 0))
        state.setdefault("setpoint", None)

    def cfg_updated(self):
        "Config has been updated"
        cg = self.cfg.get
        super().set_gains(cg("p", None), cg("i", None), cg("d", None), cg("tf", None))
        self.set_output_limits(self.cfg.get("min", None), self.cfg.get("max", None))
        self.set_offset(self.cfg.get("fct", None), self.cfg.get("off", None))

    def setpoint(self, setpoint: float):
        """
        Adjust the setpoint.
        """
        if self.state.setpoint == setpoint:
            return
        i = self.i
        if i is None:
            i = 0
        osp = self.state.setpoint

        # the default is a no-op adjustment so the factor is zero not one
        if osp is not None:
            i -= osp * self.cfg.get("factor", 0) + self.cfg.get("offset", 0)
        self.state.setpoint = nsp = setpoint
        i += nsp * self.cfg.get("factor", 0) + self.cfg.get("offset", 0)

        self.i = i

    def move_to(self, i: float, o: float, t=None):
        """
        Tell the controller that this input shall result in that output,
        by mangling its internal state as required.

        This method only works with a valid setpoint. It can be used to
        bumpless-ly hand control from one PID to another.

        Args:
            i: current process state
            o: desired controller state
        """
        if t is None:
            t = time()
        self.t = t
        if self.state.setpoint is not None:
            i -= self.state.setpoint
            self.i = o + i * self.Kp
            self.e = i

    def __call__(self, i: float, t: float | None = None) -> float:  # type:ignore[invalid-method-override]
        """
        Run a PID step.

        Args:
            i: the current output.
            t: current time, or `None` for monotonicity.
        Returns:
            the new input.
        """
        if t is None:
            t = time()
        res = super().__call__(e=self.state.setpoint - i, t=t)
        self._update_state()
        return res

    def integrate(self, i: float, t=None) -> tuple[float, float, float]:  # type:ignore[call-non-callable]
        """
        Run a PID step.

        Args:
            i: the current output.
            t: current time, or `None` for monotonicity.
        Returns:
            the p,i,d control variables.
            Call :meth:`PID.sum` to get the new input.
        """
        if t is None:
            t = time()
        res = super().integrate(e=self.state.setpoint - i, t=t)
        self._update_state()
        return res

    def _update_state(self):
        _t, e, i = self.get_state()
        self.state.e = e
        self.state.i = i



"""
A hacked-up copy of some parts of `moat.util`.
"""

from __future__ import annotations

from ._proxy import (  # noqa:F401
    DProxy,
    NoProxyError,
    Proxy,
    _CProxy,
    _RProxy,
    as_proxy,
    drop_proxy,
    get_proxy,
    name2obj,
    obj2name,
)

__all__ = [
    "DProxy",
    "NoProxyError",
    "Proxy",
    "as_proxy",
    "drop_proxy",
    "get_proxy",
    "name2obj",
    "obj2name",
    "unwrap_obj",
    "wrap_obj",
]


def wrap_obj(obj, name=None):
    "Serialize an object"
    if name is None:
        name = obj2name(type(obj))
    if isinstance(obj, Exception):
        return name, (obj.args if hasattr(obj, "args") else [])
    try:
        p = obj.__dict__
    except AttributeError:
        p = {}
        for n in dir(obj):
            if n.startswith("_"):
                continue
            p[n] = getattr(obj, n)
    return (name, (), p)


def unwrap_obj(s):
    "Deserialize an object"
    s, *d = list(s)
    st = d.pop() if d and isinstance(d[-1], dict) else {}
    try:
        p = name2obj(s)
        if hasattr(p, "__setstate__"):
            o = p(*d)
            p.__setstate__(st)
        else:
            o = p(*d, **st)
    except KeyError:
        o = DProxy(s, d, st)
    except TypeError:
        o = p(*d)
        try:
            o.__dict__.update(st)
        except AttributeError:
            for k, v in st.items():
                setattr(o, k, v)
    return o



"""
TCP multi-connection listener app.
"""

from __future__ import annotations

from moat.lib.rpc import BaseListenCmd
from moat.lib.rpc.conn.tcp import TcpIter


class Port(BaseListenCmd):
    """
    An app that accepts multiple TCP connections.
    """

    def listener(self):
        """Returns the TCP listener."""
        return TcpIter(self.cfg.get("host", "127.0.0.1"), self.cfg["port"])



"""
More common code
"""

from __future__ import annotations

from moat.util import NotGiven, attrdict
from moat.lib.codec.errors import StoppedError
from moat.lib.micro import TaskGroup, sleep_ms, ticks_diff, ticks_ms
from moat.lib.path import Path
from moat.lib.rpc import BaseCmd


class Relay(BaseCmd):
    """
    A relay is an output pin with an overriding "force" state.

    Parameters:
        pin(Path): path of the actual hardware output
        t_on(int): minimum non-forced on time
        t_off(int): minimum non-forced off time

    To prevent hardware damage, this controller can enforce the
    minimum time between relay actuations.
    """

    _delay = None
    t_last = 0
    value = None
    force = None

    doc = dict(
        _c=dict(
            _d="maybe-forced Relay",
            pin="path:Output pin, calls r+w",
            t=dict(off="int:min off-time (ms,0)", on="int:min on-time (ms,0)"),
        ),
        _d="SDP:get/set value",
        _a=[
            dict(_d="get intended value", _r="bool:state"),
            dict(_d="set intended value", _0="bool:state"),
        ],
    )

    def __init__(self, cfg):
        super().__init__(cfg)
        if not isinstance(cfg.get("pin", None), (tuple, list, Path)):
            raise ValueError("Pin not set")  # noqa:TRY004
        t = cfg.get("t", attrdict())
        self.t = [t.get("off", 0), t.get("on", 0)]

    async def reload(self):
        "reload timings"
        t = self.cfg.get("t", attrdict())
        self.t = [t.get("off", 0), t.get("on", 0)]

    async def setup(self):  # noqa:D102
        await super().setup()
        self.pin = self.root.sub_at(self.cfg["pin"])
        if await self.pin.rdy_():
            raise StoppedError("pin")
        await self.cmd_w()

    async def run(self):  # noqa:D102
        async with TaskGroup() as self.__tg:
            await super().run()

    doc_w = dict(_d="change", _0="bool:new value", f="bool|None:force?")

    async def cmd_w(self, v=None, f=NotGiven):
        """
        Change relay state.

        The state is set to @f ("force"), or @v ("value") if @f is None,
        or self.value if @v is None too.

        If you don't pass a @force argument in, the forcing state of the
        relay is not changed.
        """
        if f is NotGiven:
            f = self.force
        else:
            self.force = f

        if v is None:
            v = self.value
        else:
            self.value = v

        if f is None and self._delay is not None:
            return
        await self._set()

    async def _set(self):
        val = self.value if self.force is None else self.force
        if val is None:
            return
        p = await self.pin()
        if p == val:
            return

        if self._delay is not None:
            self._delay.cancel()
            self._delay = None
        await self.pin(val)
        t = self.t[val]
        if not t:
            return
        self._delay = await self.__tg.spawn(self._run_delay, t, _name="Rly")
        self.t_last = ticks_ms()

    async def _run_delay(self, t):
        await sleep_ms(t)
        self._delay = None
        await self._set()

    def get_sync(self):
        """
        Return the current intended state.
        """
        if self.force is not None:
            return self.force
        return self.value

    async def get(self):
        """
        Return the current intended state (async version)
        """
        return self.get_sync()

    @property
    def delayed(self):
        "flag whether the relay is delaying a change"
        return self._delay is not None

    doc_r = dict(
        _d="get state",
        _r=dict(
            v="bool:std output", f="bool:forced output", d="int:delay(ms)", p="bool:hardware state"
        ),
    )

    async def cmd_r(self):
        """
        Returns the current state, as a mapping.

        v: currently set value
        f: currently forced value
        d: delay until next change (msec) or None
        p: actual pin state
        """
        p = await self.pin.r()
        return dict(
            v=self.value,
            f=self.force,
            p=p,
            d=None if self._delay is None else ticks_diff(ticks_ms(), self.t_last),
        )

    async def cmd(self, v=None) -> bool | None:
        """
        Simple Data Protocol.
        """
        if v is None:
            return self.value
        self.value = v
        await self._set()



"""
Command tree support for MoaT commands
"""

from __future__ import annotations

from moat.lib.micro import L

from .dir import BaseSubCmd
from .layer import BaseLayerCmd

# Typing

from typing import TYPE_CHECKING, cast  # isort:skip

if TYPE_CHECKING:
    from moat.lib.micro import _TaskGroupProto
    from moat.lib.rpc import BaseConnIter
    from moat.lib.stream import BaseBuf, BaseMsg

    from collections.abc import Callable


class BaseListenOneCmd(BaseLayerCmd):
    """
    An app that runs a listener and accepts a single connection.

    Override `listener` to return it.

    TODO: this needs to be a stream layer instead: we want the
    Reliable module to be able to pick up where it left off.
    """

    def listener(self) -> BaseConnIter:
        """
        How to get new connections. Returns a BaseConnIter.

        Must be implemented in your subclass.
        """
        raise NotImplementedError

    def wrapper(self, conn) -> BaseMsg:
        """
        How to wrap the connection so that you can communicate on it.

        By default, use `~moat.lib.stream.serial_stack`.
        """
        # pylint:disable=import-outside-toplevel
        from moat.lib.stream import serial_stack  # noqa: PLC0415

        return serial_stack(conn, self.cfg)

    async def reject(self, conn: BaseBuf):
        """
        Close the connection.
        """
        # an async context should do it
        async with conn:
            pass

    async def handler(self, conn):
        """
        Process a connection
        """
        from moat.lib.rpc.cmd.msg import (  # noqa: PLC0415
            ExtCmdMsg,  # pylint:disable=import-outside-toplevel
        )

        app = ExtCmdMsg(self.cfg, self.wrapper(conn), is_server=True)
        if (
            self.app is None
            # or not await self.app.is_ready()
            or self.cfg.get("replace", True)
        ):
            if self.app is not None:
                await self.app.stop()
            app.attached(self, "_")
            self.app = app
            await self.start_app(app)
            if L:
                self.set_ready()
                await app.wait_ready()

            await app.wait_stopped()
            if self.app is app:
                self.app = None
        else:
            # close the thing
            await self.reject(conn)

    async def task(self) -> None:
        """
        Accept connections.
        """
        tg = self.tg
        if tg is None:
            raise RuntimeError("No taskgroup")
        tg = cast("_TaskGroupProto", tg)
        listener = cast("Callable[[], BaseConnIter]", self.listener)
        async with listener() as conns:
            async for conn in conns:

                async def _handle(conn=conn) -> None:
                    await self.handler(conn)

                tg.start_soon(_handle)


class BaseListenCmd(BaseSubCmd):
    """
    An app that runs a listener and connects all incoming connections
    to numbered subcommands.

    Override `listener` to return an async context manager / iterator.
    """

    seq = 1

    # no multiple inheritance for MicroPython
    listener = BaseListenOneCmd.listener
    wrapper = BaseListenOneCmd.wrapper

    async def handler(self, conn):
        """
        Process a new connection.
        """
        from moat.lib.rpc.cmd.msg import (  # noqa: PLC0415
            ExtCmdMsg,  # pylint:disable=import-outside-toplevel
        )

        wrapper = cast("Callable[[object], BaseMsg]", self.wrapper)
        conn = wrapper(conn)
        app = ExtCmdMsg(self.cfg, conn, is_server=True)
        seq = self.seq
        if seq > len(self.sub) * 3:
            seq = 10
        while seq in self.sub:
            seq += 1
        self.seq = seq + 1
        await self.attach(seq, app)
        await self.start_app(app)
        if L:
            await app.wait_ready()

        await app.wait_stopped()
        await self.detach(seq)

    async def task(self) -> None:
        """
        Accept connections.
        """
        tg = self.tg
        if tg is None:
            raise RuntimeError("No taskgroup")
        tg = cast("_TaskGroupProto", tg)
        listener = cast("Callable[[], BaseConnIter]", self.listener)
        async with listener() as conns:
            if L:
                self.set_ready()
            async for conn in conns:

                async def _handle(conn=conn) -> None:
                    await self.handler(conn)

                tg.start_soon(_handle)



"""
Non-embedded helpers, mainly for the command interpreter
"""

from __future__ import annotations

import hashlib

from ._util import TEST_MAGIC as TEST_MAGIC
from ._util import Repeater as Repeater
from ._util import Sensor as Sensor
from ._util import del_p as del_p
from ._util import get_p as get_p
from ._util import set_p as set_p
from .files import APath, copytree

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.micro.files import MoatPath


__all__ = ["Repeater", "Sensor", "del_p", "get_p", "hash256", "run_update", "set_p"]


def hash256(data):
    "Hash a chunk of bytes the way git does"
    h = hashlib.sha256()
    h.update(data)
    return h.digest()


async def _rd(f):
    "return file contents"
    async with await f.open("rb") as fd:
        return await fd.read()


async def run_update(*a, **kw):
    """
    Update a remote file system.

    The satellite contains a list of hashes for its modules.

    Thus if the source of that frozen file is identical to what we have
    now, the remote shouldn't have that file (or its .mpy derivative) in
    its file system. It might however be there as a left-over artefact from
    a previous online update. Thus we delete it.
    """
    import moat.micro._embed.lib as emb  # noqa: PLC0415

    for p in emb.__path__:
        src = APath(p)
        await _run_update(src, *a, **kw)


async def _run_update(src, dest: MoatPath, check=None, cross=None, hash_fn=None):
    # update a single _embed/lib directory

    async def drop(dst):
        """
        delete files on the satellite that didn't change between the
        version in their firmware and our current version.
        """
        # rp = dst.relative_to(emb_r)
        if dst.name == "manifest.py":
            return None

        # assume dst is relative
        sp = src.parent / dst
        # XXX we might want to ask git which files differ,
        # it's supposed to have a cache for that
        dn = str(dst)[:-3].replace("/", ".").lstrip(".")
        dn = dn.removeprefix("lib.")
        dn = dn.removesuffix(".__init__")
        if hash_fn is None:
            return False
        try:
            res = await hash_fn(dn)
            if res is None:
                return False
        except (ImportError, KeyError):
            return False
        rs = await _rd(sp)
        return res == hash256(rs)[: len(res)]

    await copytree(src, dest, check=check, drop=drop, cross=cross)



"""
Active ping
"""

from __future__ import annotations

from moat.lib.micro import sleep, wait_for
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import Msg


class _NotGiven:
    pass


class Cmd(BaseCmd):
    """
    An app to do active ping-ing: keepalive, discover dead links.

    Config:
        t: timeout
        d: delay between pings
        p: Path to call.
        s: flag whether to do streaming.
    """

    doc = dict(
        _d="echo",
        _i="Any:data",
        _o="Any:data (echoed)",
        _s=[],
        _c=dict(
            _d="echo service",
            p="path:dest for active echo-ing",
            t="int:error timeout (3 s)",
            d="int:delay(59 s)",
            s="bool:streamed?",
        ),
    )

    async def stream(self, msg: Msg):
        """
        Keepalive/liveness monitor. Echoes incoming data.
        """
        if msg.can_stream:
            async with msg.stream(*msg.args, **msg.kw) as md:
                async for m in md:
                    await msg.send(*m.args, **m.kw)
        await msg.result(*msg.args, **msg.kw)

    async def task(self):
        """
        Sender task. Counts and sends.
        """
        if (p := self.cfg.get("p", None)) is None:
            return await super().task()
        root = self.root
        if root is None:
            raise RuntimeError("Not attached")
        async with root.sub_at(p) as sub:
            if self.cfg.get("s", False):
                async with sub.stream() as msg:
                    self.set_ready()
                    await self._do(msg.send)
            else:
                self.set_ready()
                await self._do(sub)

    async def _do(self, sender):
        n = 0
        while True:
            res = await wait_for(self.cfg.get("t", 3), sender, n)
            if res != n:
                raise RuntimeError("wrong ping echo: want {n}, got {r !r}")
            n = (n + 1) if n < 9 else 1
            await sleep(self.cfg.get("d", 59))



"""
Installation and setup.
"""

from __future__ import annotations

import anyio
import logging
import os
import sys
from anyio import Path as FSPath
from anyio.to_thread import run_sync as run_sync_thread
from contextlib import suppress
from pprint import pformat
from shutil import copy as copy_fs
from shutil import copytree as copytree_fs
from shutil import rmtree

import asyncclick as click

from moat.util import NotGiven, get_part, merge, ungroup
from moat.lib.codec import get_codec
from moat.lib.micro import idle, log
from moat.lib.path import P
from moat.lib.rpc import RootCmd
from moat.micro.util import run_update
from moat.util.exec import run as run_cmd

logger = logging.getLogger(__name__)

all = ["setup", "install", "do_update", "do_copy"]  # noqa: A001


async def do_update(dst, root, cross, hfn):  # noqa: D103
    from moat.micro.files import APath, copytree  # noqa: PLC0415

    await run_update(dst / "lib", cross=cross, hash_fn=hfn)

    # do not use "/". Running micropython tests locally requires
    # all satellite paths to be relative.
    import moat.micro._embed  # noqa: PLC0415

    p = anyio.Path(moat.micro._embed.__path__[0])  # noqa:SLF001
    await copytree(APath(p / "boot.py"), root / "boot.py", cross=None)
    if not await (root / "main.py").exists():
        await copytree(APath(p / "main.py"), root / "main.py", cross=None)


async def do_copy(
    source: anyio.Path,
    dst: anyio.Path,
    dest: str | None,
    cross: str | anyio.Path | None,
    wdst: anyio.Path | None = None,
):
    """
    Copy from source to @dst/@dest.

    if @dest is `None`, append the source path behind ``/_embed/`` (if it exists).

    @cross is the path of the mpy-cross executable.
    """
    from .files import APath, copy_over  # noqa: PLC0415

    if not dest:
        dest = str(source)
        pi = dest.find("/_embed/")
        if pi > 0:
            dest = dest[pi + 8 :]
            dst /= dest
    else:
        dst /= dest
    awdst = None if wdst is None else APath(wdst)
    await copy_over(source, dst, cross=cross, wdst=awdst)


def _clean_cfg(cfg):
    # cfg = attrdict(apps=cfg["apps"])  # drop all the other stuff
    return cfg


async def setup(
    cfg,
    install: bool = False,
    source: anyio.Path | None = None,
    root: str = ".",
    dest: str = "",
    kill: bool = False,
    large: bool | None = None,
    run: bool | dict = False,
    rom: bool = False,
    reset: bool = False,
    state: str | None = None,
    config: dict | None = None,
    cross: str | anyio.Path | None = None,
    update: bool = False,
    watch: bool = False,
    main: str | anyio.Path | bool | None = False,
    config_file: str = "moat.cfg",
):
    """
    Given the serial link to a MicroPython board,
    teach it to run the MoaT loop.

    Parameters: see "moat micro setup --help".
    """
    # 	if not source:
    # 		source = anyio.Path(__file__).parent / "_embed"

    if bool(watch) + bool(run) > 1:
        raise click.UsageError("You can only use one of 'watch' or 'run'.")

    if install:
        await install_(cfg, dest=dest)
        print("Firmware installation done.", file=sys.stderr)

    from moat.lib.stream import RemoteBufAnyio  # noqa: PLC0415

    from .direct import DirectREPL  # noqa: PLC0415
    from .files import ABytes, MoatDevPath, copy_over  # noqa: PLC0415

    codec = get_codec("std-cbor")

    if cross == "-":
        cross = None
    elif cross is not None:
        cross = str(cross)

    if kill:
        async with (
            RootCmd(cfg) as dsp,
            dsp.sub_at(cfg.remote) as sd,
            RemoteBufAnyio(sd) as ser,
            DirectREPL(ser) as repl,
        ):
            dst = MoatDevPath(root).connect_repl(repl)
            await repl.reset()
        await anyio.sleep(2)

    need_run = False

    if main == "-":
        main = None
    elif main is None and "install" in cfg:
        main = cfg.install.get("main", None)
        if main is NotGiven:
            main = None
        else:
            import moat.micro._embed  # noqa: PLC0415

            main = cfg.install.get("main", anyio.Path(moat.micro._embed.__path__[0]) / "main.py")  # noqa:SLF001

    # The following dance is necessary because a reset may or may not kill
    # the whole stack. Fixing this, i.e. making individual apps fault
    # tolerant, is somewhere on the TODO list.

    async def part_two():
        nonlocal need_run
        if run or watch:
            if run and not reset:
                o, e = await repl.exec_raw(
                    f"import moat; moat.go(state={state!r})",
                    timeout=None if watch else 30,
                )
                if o:
                    print(o)
                if e:
                    print("ERROR", file=sys.stderr)
                    print(e, file=sys.stderr)
                    sys.exit(1)

            if watch:
                while True:
                    d = await ser.receive()
                    sys.stderr.buffer.write(d)
                    sys.stderr.buffer.flush()

            if run:
                log("Reloading.")
                merge(dsp.cfg, run)
                await dsp.reload()
                log("Running.")
                cons = b""
                while True:
                    try:
                        with anyio.fail_after(3):
                            (m,) = await dsp.cmd(P("s.crd"))
                        cons += m
                        if cons.endswith(b"\nOK\x04\x04>"):
                            break
                    except TimeoutError:
                        break
                if cons:
                    log("Console:\n%s", cons.decode("utf-8", errors="replace"))
                async with dsp.sub_at(dsp.cfg.remote2) as sr:
                    try:
                        print(await sr.dir_())
                        await sr.s.ping()
                    except Exception as exc:
                        logger.error("PROBLEM %r", exc)
                    need_run = False
                    log("RUNNING.")
                    await idle()

    try:
        async with (
            RootCmd(cfg) as dsp,
            dsp.sub_at(cfg.remote) as sd,
            RemoteBufAnyio(sd) as ser,
            DirectREPL(ser) as repl,
        ):
            dst = MoatDevPath(root).connect_repl(repl)
            if source:
                if rom:
                    from .romfs import make_romfs, write_romfs  # noqa:PLC0415

                    async with anyio.TemporaryDirectory() as tf:
                        wdst = anyio.Path(tf)
                        await do_copy(source, dst, dest, cross, wdst=wdst)
                        rom_data = await make_romfs(wdst, cross)
                        await write_romfs(sd, rom_data)
                else:
                    await do_copy(source, dst, dest, cross)
            if state and not watch:
                await repl.exec(f"f=open('moat.state','w'); f.write({state!r}); f.close(); del f")
            if large is True:
                await repl.exec("f=open('moat.lrg','w'); f.close()", quiet=True)
            elif large is False:
                await repl.exec("import os; os.unlink('moat.lrg')", quiet=True)

            if config:
                config = _clean_cfg(config)
                logger.debug("Config:\n%s", pformat(config))
                f = ABytes(name=config_file, data=codec.encode(config))
                await copy_over(f, MoatDevPath(config_file).connect_repl(repl))

            if main:
                from moat.micro.files import APath, copytree  # noqa: PLC0415

                await copytree(
                    APath(anyio.Path(str(main))),
                    MoatDevPath(".").connect_repl(repl).joinpath("main.py"),
                    cross=None,
                )

            if update:

                async def hfn(p):
                    res = await repl.exec(
                        f"import _hash; print(repr(_hash.hash[{p!r}])); del _hash",
                        quiet=True,
                    )
                    return eval(res)

                await do_update(dst, MoatDevPath(".").connect_repl(repl), cross, hfn)

            need_run = True
            if reset:
                await repl.soft_reset(run_main=bool(run))
                # reset with run_main set should boot into MoaT

            await part_two()

    except Exception as exc:
        if not need_run:
            raise

        logger.warning(
            "Reset: caused %r", exc, exc_info=exc if logger.isEnabledFor(logging.DEBUG) else None
        )

        await anyio.sleep(3)
        async with (
            RootCmd(cfg) as dsp,
            dsp.sub_at(cfg.remote) as sd,
        ):
            log("Running after reset.")
            await part_two()


def find_p(prog: str):
    """
    Find an executable in the system path.
    """
    for p in os.environ["PATH"].split(os.pathsep):
        try:
            pp = p + os.sep + prog
            os.stat(pp)
        except FileNotFoundError:
            continue
        else:
            return pp

    return None


async def install_(cfg, dest: str | anyio.Path | None = None):
    """
    Install our version of MicroPython to a device.
    """
    device = cfg.install.port
    board = cfg.install.get("board", None)
    variant = cfg.install.get("variant", None)

    mydir = FSPath(__file__).parent.parent.parent
    mpydir = mydir / "ext" / "micropython"
    if not await mpydir.is_dir():
        mpydir = FSPath(cfg.install.micropython)
    portdir = mpydir / "ports" / device

    boardp = f"{board or 'generic'}" + (f"-{variant}" if variant else "")

    board_dir = FSPath(__file__).parent / "_embed" / "boards" / device / board
    if await board_dir.exists():
        board_arg = f"{device}_moat{os.getpid()}"
        board_tmp = portdir / "boards" / board_arg
    else:
        board_tmp = None
        board_dir = board_arg = boardp
        if not await (portdir / board_dir).exists():
            board_dir = None
    board_arg = str(anyio.Path("boards") / board_arg)

    try:
        raise OSError("doesn't work")
        build_dir = FSPath(cfg.install.build)
        if not await build_dir.exists():
            await build_dir.mkdir()

        buildp = f"build-{os.getpid()}"
        build_tmp = portdir / buildp
    except OSError:
        buildp = f"build-{boardp}"
        build_dir = buildp
        build_tmp = None

    if isinstance(cfg.install.serial, str):
        port = anyio.Path(cfg.install.serial)
    else:
        try:
            port = anyio.Path(get_part(cfg, cfg.install.serial))
        except AttributeError:
            port = None
    try:
        rate = cfg.install.rate
    except AttributeError:
        rate = None

    if device == "rp2" and dest is None:
        raise ValueError("Installing to Raspberry Pi Pico requires a 'dest' directory")

    args = []

    if device == "esp32":
        idf = find_p("idf.py")
        if idf is None:
            if "ESP" not in os.environ:
                raise click.UsageError(
                    "'idf.py' not found: Source $ESP/export.sh and try again.",
                )
            idf = os.environ["ESP"] + os.sep + "idf.py"

    if device in ("esp32", "esp8266"):
        goal = "deploy"
        args.append("ESPTOOL=esptool")

    else:
        goal = "all"
        # if board is None:
        #     board = "rpi_pico"

    import moat.micro._embed._tag as m  # noqa: PLC0415

    manifest = m.__file__.replace("_tag", "manifest")

    env = dict(os.environ)
    env["PYTHONPATH"] = str(await anyio.Path.absolute(mydir))

    try:
        if build_tmp:
            await (build_tmp).symlink_to(build_dir)
        if board_tmp:
            await run_sync_thread(copytree_fs, str(board_dir), str(board_tmp))

        await run_cmd(
            "make",
            "-j",
            cwd=str(mpydir / "mpy-cross"),
            echo=True,
        )
        await run_cmd(
            "make",
            "-j",
            "PORT=" + str(port),
            "BAUD=" + str(rate),
            "BOARD=" + board,
            "BOARD_DIR=" + board_arg,
            "BUILD=" + buildp,
            "FROZEN_MANIFEST=" + manifest,
            *args,
            goal,
            cwd=portdir,
            env=env,
            echo=True,
        )
    finally:
        with anyio.move_on_after(2, shield=True):
            if build_tmp:
                with suppress(OSError), ungroup:
                    await (portdir / build_tmp).unlink()
            if board_tmp:
                with suppress(OSError), ungroup:
                    await run_sync_thread(rmtree, str(board_tmp))

    if device == "rp2":
        if isinstance(dest, str):
            dest = anyio.Path(dest)
        assert dest is not None
        df = dest / "INFO_UF2.TXT"
        if not await df.exists():
            print(f"Waiting for RPI in {dest} ", end="")
            sys.stdout.flush()
            while not await df.exists():  # noqa:ASYNC110
                # not going to import inotify just for this
                await anyio.sleep(1)
            print("… found.")

        await run_sync_thread(
            copy_fs,
            portdir / "build-RPI_PICO/" / "firmware.uf2",
            dest,
        )

    if port is not None and not await port.exists():
        print(f"Waiting for RPI in {dest} ", end="")
        sys.stdout.flush()
        while not await port.exists():  # noqa:ASYNC110
            # not going to import inotify just for this
            await anyio.sleep(1)
        print("… found.")

    await setup(
        cfg,
        run=True,
        update=True,
        state="once",
        cross=str(mpydir / "mpy-cross" / "build" / "mpy-cross"),
    )



"""
PID controller.
"""

from __future__ import annotations

from moat.lib.broadcast import Broadcaster
from moat.lib.micro import AC_use, TaskGroup
from moat.lib.path import Path
from moat.lib.pid import CPID
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import Msg, MsgSender

_state_d = dict(t="int:last time", e="float:error", i="float:integral")
_pid_d = dict(p="float:P", i="float:I", d="float:D", tf="float:Filter D")


class PID(BaseCmd):
    """
    A PID controller periodically (we hope) receives an input value, and
    returns an output to adjust the input.

    - min, max: output value boundaries
    - p: proportional gain
    - i: integral gain
    - d: differential gain
    - tf: first-order filter time constant for the differential, in seconds
    - state: path to our state storage
    - set: initial goal (default zero)

    t should be larger than the interval between inputs.
    """

    pid: CPID
    val_in: float | None = None
    split: tuple[float, float, float] | None = None
    state_path: Path | None = None
    state_rtc: MsgSender | None = None

    doc = dict(
        _c=dict(
            _d="PID control",
            state=dict(_d="initial state", **_state_d),
            factor="float:setpoint adj factor (0)",
            offset="float:setpoint adj offset (0)",
            setpoint="float:initial goal",
            **_pid_d,
        )
    )

    def __init__(self, cfg):
        super().__init__(cfg)
        self.pid = CPID(cfg)
        self.state_path = cfg.get("state", None)
        self.pid.setpoint(cfg.get("set", 0))

        self._set_bc = Broadcaster(1)

    async def setup(self):
        "retrieve state"
        await super().setup()
        await AC_use(self, self._set_bc)
        if self.state_path is not None:
            sr = self.root.sub_at(self.state_path)
            await sr.rdy_()
            try:
                st = await sr.r()
            except KeyError:
                pass
            else:
                setpoint = st.pop("setpoint", None)
                if setpoint is not None:
                    self.pid.setpoint(setpoint)
                self.pid.set_state(**st)
            self.state_rtc = sr

    doc_sw = dict(_d="update state", **_state_d)

    def cmd_sw(self, t: int | None, e: float | None, i: float | None, **_kw):
        "Update the PID state."
        self.pid.set_state(t, e, i)

    doc_w = dict(_d="step", _0="float:current value", _r="float:new output")

    async def cmd_w(self, val: float, t: int | None = None) -> float:
        """
        Run a PID step.
        """
        self.val_in = val
        self.split = s = self.pid.integrate(val, t=t)
        if (sr := self.state_rtc) is None and self.state_path is not None:
            self.state_rtc = sr = self.root.sub_at(self.state_path)
            await sr.rdy_()
        if sr is not None:
            await sr.cmd(Path.build(("w",)), self.pid.state, fs=False)
        return self.pid.sum(s)

    doc_sp = dict(
        _d="setpoint",
        _a=[
            dict(_d="set", _0="float:new setpoint"),
            dict(_d="read", _r="float:current setpoint"),
        ],
    )

    async def cmd_sp(self, sp: float | None = None):
        "Sets/Returns the current setpoint"
        if sp is None:
            return self.pid.state.setpoint
        self.pid.setpoint(sp)
        self._set_bc(sp)

    doc_s = dict(
        _d="read state",
        _r=dict(
            state=_state_d,
            i="float:last input",
            o="float:last output",
            split="list[float]:p-i-d output",
            fct=["float:factor", "float:offset"],
        ),
        **_state_d,
    )

    async def stream_sp(self, msg: Msg):
        "Sets/Returns the current setpoint"
        async with msg.stream() as ms, TaskGroup() as tg:

            @tg.start_soon
            async def _sp_evt():
                async with self._set_bc.reader(1) as spr:
                    async for sp in spr:
                        await ms.send(sp)

            async for m in ms:
                self.pid.setpoint(m[0])
                self._set_bc(m[0])

    async def cmd_s(self, **kw):
        """
        Sets/Returns the current state.
        """
        if kw:
            self.pid.set_state(**kw)
        res = dict(
            state=self.pid.state,
            i=self.val_in,
            split=self.split,
            fct=self.pid.get_offset(),
            gain=self.pid.get_gains(),
        )
        if self.split:
            res["o"] = self.pid.sum(self.split)
        return res

    async def reload(self):
        "reload me"
        await super().reload()
        self.pid.cfg_updated()
        self.state_rtc = None
        self.state_path = self.cfg.get("state", None)



"""
Implements an alarm aggregator.

Alarms consist of a condition object, a path, and some data.
They can be raised, updated, or lowered.

The aggregator's job is to collect the alarms so that the server can poll
a single object.

Alarm conditions are subclasses of exceptions.
"""

from __future__ import annotations

from moat.util import merge
from moat.lib.micro import AC_use, Event, Queue, TaskGroup, WouldBlock
from moat.lib.path import Path
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.micro import _TaskGroupProto
    from moat.lib.rpc import Msg

    from collections.abc import Iterator


__all__ = [
    "Alert",
    "AlertHandler",
]


class Alert(Exception):
    """Alert wrapper. The path isn't stored in it."""

    def __init__(self, data):
        self.data = data


class AlertIter:
    q: Queue
    xal: Iterator[tuple[tuple[type[Alert], Path], Alert]]

    def __init__(self, ah: AlertHandler, s: bool | None):
        self.ah = ah
        self.s = s

    async def __aenter__(self):
        if hasattr(self, "q"):
            raise RuntimeError("already on")
        self.q = q = Queue(10)
        self.ah.q.add(q)
        self.xal = iter(() if self.s is False else list(self.ah.alarms.items()))
        return self

    async def __aexit__(self, *tb):
        self.ah.q.discard(self.q)
        del self.q

    def __aiter__(self):
        return self

    async def __anext__(self):
        a = None
        try:
            k, al = next(self.xal)
        except AttributeError:
            pass
        except StopIteration:
            del self.xal
            if self.s:
                raise StopAsyncIteration from None
        else:
            a, p = k
            d = al.data

        if a is None:
            try:
                q = self.q
                if q is None:
                    raise RuntimeError("not started")
                a, p, d = await q.get()
            except EOFError:
                raise StopAsyncIteration  # noqa:B904,RUF100
        res = {"a": a, "p": p}
        if d is not None:
            res["d"] = d
        return res


class AlertHandler(BaseCmd):
    """
    Collect open alerts.

    This helper class keeps track of open alerts and lets multiple clients
    receive them.

    This command can monitor + redistribute alerts from satellites::

        apps:
          al: link.Alert
          r: …
        al:
          mon:
          - rem: !P r
            al: !P a
        r:
          cfg:
            a: link.Alert

    The rem/al split is used because incoming alert paths are prefixed with
    the "rem" path so that the destination is known and unique.
    """

    q: set[Queue]
    tg: _TaskGroupProto

    def __init__(self, cfg):
        super().__init__(cfg)
        self.alarms: dict[tuple[type[Alert], Path], Alert] = {}
        self.q = set()
        self._mon = {}

    async def setup(self) -> None:  # noqa:D102
        await super().setup()
        await self._start_mon()

    async def _rdr(self, p: dict, evt: Event) -> None:
        rem = p["rem"]
        al = p["al"]
        async with self.root.cmd(Path.build(rem + al + ("r",))).stream_in() as it:
            evt.set()
            async for res in it:
                await self.cmd_w(a=res["a"], p=rem + res["p"], d=res.get("d", None))

    async def _start_mon(self) -> None:
        if (pl := self.cfg.get("mon", None)) is not None:
            if getattr(self, "tg", None) is None:
                self.tg = await AC_use(self, TaskGroup())
            tg = self.tg

            evs = []
            for k, v in pl.items():
                evt = Event()

                async def _run(v=v, evt=evt) -> None:
                    await self._rdr(v, evt)

                self._mon[k] = await tg.spawn(_run)
                evs.append(evt)
            for e in evs:
                await e.wait()

        else:
            pl = {}

        for k, v in list(self._mon.items()):
            if k not in pl:
                v.cancel()
                del self._mon[k]

    async def reload(self) -> None:  # noqa:D102
        await self._start_mon()

    doc_r = dict(_d="Stream curren alerts", s="bool:stop(default:wait)", _o="alert")

    async def stream_r(self, msg: Msg) -> None:
        """read open alarms.

        If @s ("static") is True, send a snapshot.
        Otherwise iterate on new alerts.
        """
        async with msg.stream_out(), AlertIter(self, msg.get("s", False)) as alit:
            async for al in alit:
                await msg.send(**al)

    doc_w = dict(_d="set alert", _0="type:class", _1="path", d="any:data, clears if missing")

    async def cmd_w(self, a: type[Alert], p: Path, d: dict | None = None) -> None:
        """
        Set an alert.

        @a: alarm class (a type, *not* an object)
        @p: path to the faulting object
        @d: data describing the current state

        If @d is `None`, the alert is cleared.
        """
        k = (a, p)
        al = self.alarms.get(k, None)
        if al is None:
            if d is None:
                return  # nothing to do
            self.alarms[k] = a(d)
        else:
            if d is not None:
                merge(al.data, d)
            else:
                del self.alarms[k]

        for q in list(self.q):
            try:
                q.put_nowait((a, p, d))
            except WouldBlock:
                self.q.discard(q)
                q.close_sender()

    doc_cl = dict(_d="close iters")

    async def cmd_cl(self):
        """
        Close all iterators of this alert handler.

        Used mainly for testing.
        """
        q, self.q = self.q, set()
        while q:
            q.pop().close_sender()



"""
This code is used by MicroPython to set up the test dispatcher.

TODO: replace this with the standard "main.py".
"""

from __future__ import annotations

import sys

sys.path.insert(0, "./stdlib")
sys.path.insert(0, ".")

import moat

moat.go(cmd=False)



"""
Basic handler for iterating incoming Moat connections.
"""

from __future__ import annotations

from moat.util import Queue
from moat.lib.micro import ACM, AC_exit, Event, L, TaskGroup

# typing
from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from moat.lib.stream import BaseConn

    from collections.abc import Awaitable
    from typing import Never


class BaseConnIter:
    """
    Iterate incoming connections.

    You need to override the "accept" method.
    """

    def __init__(self):
        self.q = Queue(1)
        self.evt = Event()

    async def __aenter__(self):
        self.tg = await ACM(self)(TaskGroup())
        try:
            self.tg.start_soon(self.accept)
            return self
        except BaseException as exc:
            await AC_exit(self, type(exc), exc, getattr(exc, "__traceback__", None))
            raise

    async def __aexit__(self, *exc):
        await AC_exit(self, *exc)

    if L:

        def set_ready(self) -> None:
            "signals that the socket-or-whatever accepts connections"
            self.evt.set()

        def is_ready(self) -> Awaitable:
            "wait for the socket-or-whatever to accept connections"
            return self.evt.wait()

    def add_conn(self, c: BaseConn) -> Awaitable:
        "queues the connection for starting a task"
        return self.q.put(c)

    async def accept(self) -> Never:
        """
        Background task to accept incoming connections.

        Call ``await self.add_conn(conn)`` for each connection.

        Call ``self.ready()`` as soon as the socket-or-whatever is ready to accept links.
        """
        raise NotImplementedError

    def __aiter__(self):
        return self

    def __anext__(self) -> Awaitable[BaseConn]:
        return self.q.get()



"""
Access a satellite's Flash file system.
"""

from __future__ import annotations

import vfs

from moat.lib.rpc import BaseCmd


class Dev:
    "Mostly-byte-level abstraction for a ROMFS device"

    _wr = None

    doc = dict(_c=dict(_d="romfs abstraction, TODO"))

    def __init__(self, n):
        self.dev = vfs.rom_ioctl(2, n)
        self.buf = bytearray(self.dev.ioctl(5, 0))
        self.pos = 0

    @property
    def cur_blk(self):
        "current block number"
        return self.pos // len(self.buf)

    @property
    def cur_off(self):
        "offset in current block"
        return self.pos % len(self.buf)

    @property
    def buf_n(self):
        "device's number of buffers"
        return self.dev.ioctl(4, 0)

    @property
    def buf_sz(self):
        "device's buffer size"
        return len(self.buf)

    def seek(self, n: int):
        """go to @n, clears r/w flag"""
        if n == self.pos:
            return
        if n % len(self.buf):
            raise ValueError("block boundary only")
        if self._wr and self.pos % len(self.buf):
            raise RuntimeError("buffered write")
        self.pos = n
        self._wr = None

    def write(self, data: bytes):
        """write to the buffer"""
        if self._wr is False:
            raise ValueError("r/w")
        self._wr = True
        ld = len(data)
        if self.pos % len(self.buf) + ld > len(self.buf):
            raise ValueError("crosses block boundary")
        self.buf[self.cur_off : self.cur_off + ld] = data
        self.pos += ld
        if self.cur_off == 0:
            self.wrb()

    def wrb(self):
        """write the current buffer"""
        lb = len(self.buf)
        blk = self.cur_off // lb - (self.cur_off % lb == 0)
        self.dev.writeblocks(blk, self.buf)  # no offset given: auto-erase

    def read(self, n: int):
        """fetch @n bytes from the buffer"""
        if self._wr is False:
            raise ValueError("r/w")
        self._wr = False
        if self.cur_off % len(self.buf) == 0:
            self.rdb()
        np = self.pos % len(self.buf)
        if np + n > len(self.buf):
            raise ValueError("crosses block boundary")
        self.pos += n
        return memoryview(self.buf)[np : np + n]

    def rdb(self):
        """read the next buffer"""
        lb = len(self.buf)
        self.dev.readblocks(self.cur_off // lb, self.buf)

    def close(self):
        """flush the buffer"""
        if self._wr:
            self.wrb()
        self.seek(0)


class Cmd(BaseCmd):
    """
    ROM File system access.

    Set "root" to the file system path this command should apply to.
    """

    _dev_cache = None

    def __init__(self, cfg):
        super().__init__(cfg)
        self._dev_cache = dict()
        self._dev_pos = dict()

    def dev(self, n):  # noqa: D102
        try:
            return self._dev_cache[n].dev
        except KeyError:
            try:
                vfs.umount("/rom")
            except Exception:  # noqa:S110
                pass
            self._dev_cache[n] = dev = vfs.rom_ioctl(2, n)
            return dev

    doc_n = dict(_d="#ROMs", _r="int:n")

    async def cmd_n(self):
        "return the number of ROM filesystems"
        return vfs.rom_ioctl(1)

    doc_stat = dict(_d="get device sizes", _0="int:n", _r0="int:nblocks", _r1="int:blocksize")

    async def cmd_stat(self, dev: int):
        "device data"
        d = self.dev(dev)
        return d.buf_n, d.buf_sz

    doc_cl = dict(_d="close device file", _0="int:n")

    async def cmd_cl(self, dev: int):
        "'close' device N"
        dev = self._dev_cache.pop(dev)
        dev.close()
        return

    doc_rd = dict(_d="read FS", _0="int:dev", _1="int:offset", n="int:length")

    async def cmd_rd(self, dev: int = 0, o: int = 0, n: int = 64):
        "read @n bytes from @f at offset @o"
        dev = self.dev(dev)
        dev.seek(o)
        return dev.read(n)

    async def stream_rd(self, msg):  # noqa: D102
        dev = self.dev(msg[0])
        off = msg[1] if len(msg) > 1 else 0
        blk = msg[2] if len(msg) > 2 else 64
        nblk = dev.buf_sz * dev.buf_n
        nblk = (nblk + blk - 1) // blk
        off  # noqa:B018
        async with msg.stream_out() as st:
            while nblk > 0:
                nblk -= 1
                await st.send(dev.read(blk))
        dev.close()

    doc_wr = dict(_d="write FS", _0="int:dev", _1="int:offset", d="bytes:data")

    async def cmd_wr(self, dev, o=0, d=None):
        "write @d to @d at offset @o"
        dev = self.dev(dev)
        dev.seek(o)
        return dev.write(d)

    async def stream_wr(self, msg):  # noqa: D102
        dev = self.dev(msg[0])
        off = msg[1] if len(msg) > 1 else 0
        dev.seek(off)
        async with msg.stream_in() as st:
            async for (m,) in st:
                dev.write(m)
        dev.close()



"""
Serial link app for MoaT message exchange.
"""

from __future__ import annotations

from moat.lib.micro import AC_use
from moat.lib.rpc.cmd.msg import BaseCmdMsg
from moat.lib.stream import serial_stack
from moat.micro.app._doc import _link_d, _log_d, _mode_d

from ._util import get_serial


class Link(BaseCmdMsg):
    """r/w: exchange MoaT messages, possibly framed."""

    doc = dict(
        _c=dict(
            _d="message serial data",
            port="str:Port or 'USB'",
            mode=_mode_d,
            link=_link_d,
            **_log_d,
        )
    )

    async def stream(self):
        """Returns the console-stack-wrapped serial stream."""
        return await AC_use(self, serial_stack(get_serial(self.cfg), self.cfg))



"""
RTC access
"""

from __future__ import annotations

from moat.util import NotGiven, enc_part, get_part, merge, set_part
from moat.lib.rpc import BaseCmd
from moat.micro.rtc import RTC
from moat.util.exc import ExpKeyError

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.path import PathElem
    from moat.lib.rpc import Msg


class _NotGiven:
    pass


class Cmd(BaseCmd):
    """
    Subsystem to handle RTC data.

    This app serves access to the config snippet(s) stored in the system's
    RTC chip.
    """

    doc = dict(_c=dict(_d="RTC access"))

    async def cmd(self, k=_NotGiven, v=_NotGiven):
        """
        Direct R/W access to RTC.

        Args:
            k: Key to access.
            v: Value to set. If not given, returns current value.

        This function does not list keys. Use :meth:`stream` for that.

        The name can also be passed as next-to-last path element.
        (The last is `w`.)
        """
        if k is _NotGiven:
            return await RTC.keys()
        if v is _NotGiven:
            return await RTC.get(k)
        else:
            await RTC.set(k, v)

    doc_r = dict(_d="read data", p="path", _r="parts")

    def __init__(self, cfg: dict):
        super().__init__(cfg)
        self._data = {}

    async def stream_r(self, msg: Msg):
        """
        Read (part of) the RTC data area.

        As data are frequently too clunky to transmit in one go, this code
        interrogates the state step-by-step.

        Args:
            0(str): the config entry.
            1(Path): the path to access within the entry.

        If the accessed item is a dict, return data consists of a dict
        (simple keys and their values) and a list (keys for complex
        values).

        Same for a list.

        RTC.get() is called when the path is empty, otherwise the cached
        value is used.  Keyword args are passed to it.
        """
        # This is a stream command because it returns the enc_part output
        # as multi-element return values.
        try:
            n = msg.kw.pop("n")
        except KeyError:
            n = msg[0]
        p = msg.kw.pop("p", ())

        try:
            if not p:
                self._data[n] = res = await RTC.get(n, **msg.kw)
            else:
                res = get_part(self._data[n], p)
            res = enc_part(res)
            if isinstance(res, (list, tuple)):
                await msg.result(*res)
            else:
                await msg.result(res)
        except KeyError as exc:
            raise ExpKeyError(*exc.args) from None

    doc_w = dict(_d="write data", _0="str:name", _1="any:data")

    async def stream_w(self, msg: Msg):
        """
        Write to RTC.

        Args:
            name(str): Name of the entry.
            data(Any): Data to store. `NotGiven` deletes.

        The name can also be passed as next-to-last path element.
        (The last is `w`.)
        """
        try:
            n = msg.kw.pop("n")
        except KeyError:
            n = msg[0]
            off = 1
        else:
            off = 0
        try:
            d = msg.kw.pop("d")
        except KeyError:
            d = msg[off]

        return await RTC.set(n, d, **msg.kw)

    doc_c = dict(_d="write cached data", _0="str:name", _1="path:pos", d="any:deletes if missing")

    async def stream_c(self, msg: Msg):
        """
        Write to the cached data.

        As data are sometimes too large to transmit in one go, this code
        updates the state step-by-step.

        Args:
            0(str): the config entry.
            1(Path): the path to access within the entry.
            d(Any): data.

        If the path is empty, the cached data is replaced/dropped depending
        on whether *d* is `NotGiven`.

        This method does not write to RTC. Use :meth:`cmd_x` for that.
        """
        n = msg.kw.pop("n")
        p = msg.kw.pop("p", ())
        d = msg.kw.pop("d", NotGiven)
        keep = msg.kw.pop("keep", False)
        if not p:
            if d is NotGiven:
                self._data.pop(n, None)
            else:
                self._data[n] = d
        else:
            set_part(self._data.setdefault(n, {}), p, d, keep=keep)

    doc_x = dict(_d="activate data")

    async def cmd_x(self, n: str, **kw):
        """
        Write the current cache entry with this name to RTC.

        If no such entry exists, the datum is deleted from RTC.

        This command replaces the RTC config if the name "rtc" is written.
        It updates (not replaces) the global config if "cfg" is written.

        """
        d = self._data.pop(n, NotGiven)
        res = await RTC.set(n, d, **kw)
        if d is not NotGiven:
            if n == "rtc":
                RTC.init(d)
            elif n == "cfg":
                merge(self.root.cfg, d, replace=True)
                await self.root.reload()
        return res

    async def stream(self, msg: Msg):
        """
        Get all data.

        Arguments:
            keys: flag whether to get only keys. If `False`, return
                  ``(key,value)`` pairs. The default is `True`.
        """
        keys = msg.kw.get("keys", True)
        fs = msg.kw.get("fs", None)
        async with msg.stream() as ms:
            for k in await RTC.keys(fs=fs):
                if not keys:
                    k = (k, await RTC.get(k, fs=fs))  # noqa:PLW2901
                await ms.send(k)

    async def handle(self, msg: Msg, rcmd: list[PathElem]):
        """
        Handler override for adding the name to the path.

        rcmd is reversed, so rtc.cfg.w arrives as ["w", "cfg"]
        """
        if len(rcmd) == 2 and "n" not in msg.kw:
            n_ = rcmd[1]
            c_ = rcmd[0]
            if (
                isinstance(n_, str)
                and n_
                and n_[-1] != "_"
                and isinstance(c_, str)
                and c_ in "rwcx"
            ):
                msg.kw["n"] = rcmd.pop()
        await super().handle(msg, rcmd)



"""
This library contains an advanced [PID controller](https://en.wikipedia.org/wiki/Proportional%E2%80%93integral%E2%80%93derivative_controller).

The `PID` class implements a controller with the error term as its input.
The `CPID` version uses a setpoint variable and accepts the
actual system state as its input.

Both variants implement
- timing independence
- a first-order filter on the derivaive term
- wind-up protection
- introspection
- saving and restoring the controller's state

TODO:
- bump-less parameter changes
"""

from __future__ import annotations

from ._impl import CPID as CPID
from ._impl import PID as PID
from ._impl import PID_TC as PID_TC

__all__ = ["CPID", "PID"]



"""
exponential moving average
"""

from __future__ import annotations

from math import exp

from moat.lib.micro import Event, ticks_diff, ticks_ms
from moat.lib.rpc import BaseCmd, NoStream

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import Msg


class Average(BaseCmd):
    """
    This is an exponential moving average.

    Iterating it yields a new value whenever the input changes.

    Config:
      t: filter reaction time constant, in seconds. Must be given.
      init: initial value
      age: age of the initial value, in seconds; higher = less weight
    """

    doc = dict(
        _c=dict(
            _d="Exponential average",
            init="float:initial value",
            t="float:time const (s)",
            age="float:init age",
        )
    )

    flag: Event

    def __init__(self, cfg):
        super().__init__(cfg)
        if "t" not in self.cfg:
            raise ValueError("Average: need react time constant")
        self._value = cfg.get("init", None)
        self._t = None if self._value is None else ticks_ms() - cfg.get("age", self.cfg["t"])
        self.flag = Event()

    def in_value(self, val, t=None):
        "update value"
        tn = ticks_ms()
        if t is None:
            t = ticks_diff(tn, self._t) if self._t is not None else 1
        if self._value is not None:
            val = self._value + (1 - exp(-t / 1000 / self.cfg["t"])) * (val - self._value)

        self._value = val
        self._t = tn
        self.flag.set()
        self.flag = Event()

    doc_r = dict(
        _d="read", t="int:last timestamp (systime)", _r="float:current avg", _s=True, _o=True
    )

    async def stream_r(self, msg: Msg):
        "read. Wait for change if timestamp didn't change"
        if self._t is None or msg.get("t", None) == self._t:
            await self.flag.wait()
        if msg.can_stream:
            async with msg.stream_out(t=self._t) as m:
                while True:
                    try:
                        await m.send(self._value)
                    except NoStream:
                        break
                    await self.flag.wait()
        else:
            await msg.result(self._value, t=self._t)

    doc_w = dict(_d="write", _0="float:update", _t="float:Timestamp(ms)", _s=True, _i=True)

    async def stream_w(self, msg):
        "update"
        if msg.can_stream:
            async with msg.stream_in() as mon:
                async for m in mon:
                    self.in_value(m[0], m.get("t", None))
                    await mon.send(self._value, t=self._t)
        else:
            self.in_value(msg[0], msg.get("t", None))
        await msg.result(self._value, t=self._t)



"""
This module implements a simple framing scheme for serially-transmitted
data. Frames use an optional fixed start byte, a intra-frame time maximum
to protect against data loss, and a CRC16 (polynomial 0xBAAD, optimal for
small-ish frames with a large number of bit errors).

The length is transmitted as one or two characters, depending on its
maximum, in order to keep packets small yet not restrict size too much.

A C implementation of this scheme is also available. The C version supports
streaming; this Python implementation does not.
"""

from __future__ import annotations

__all__ = ["CRC16", "FRAME_START", "SerialPacker"]

try:
    import utime as time
except ImportError:
    # CPython
    import time

    def ticks_ms():
        """System ticks, in msec"""
        return time.clock_gettime_ns(time.CLOCK_MONOTONIC_RAW) / 1000000

    def ticks_diff(a, b):
        """Difference between ticks, non-wrapping"""
        return a - b

else:
    # MicroPython
    ticks_ms = time.ticks_ms
    ticks_diff = time.ticks_diff

FRAME_START = 0xFA


class CRC16:
    """
    A simple CRC routine.

    Polynomial: 0xBAAD, no inversion or similar tricks
    """

    # cd moat-bus/python/moatbus
    # python3 ./crc.py -b16 -p0xBAAD -d4
    len = 2

    # Use a nibble-based table; it's fast enough, but doesn't need a
    # kbyte of memory (MicroPython doesn't have arrays).
    table = (
        0x0000,
        0x64A8,
        0xC950,
        0xADF8,
        0xE7FB,
        0x8353,
        0x2EAB,
        0x4A03,
        0xBAAD,
        0xDE05,
        0x73FD,
        0x1755,
        0x5D56,
        0x39FE,
        0x9406,
        0xF0AE,
    )

    def __init__(self):
        self.crc = 0

    def feed(self, byte):
        """add a byte to the CRC"""
        crc = self.crc
        crc = self.table[((byte >> 4) ^ crc) & 0xF] ^ (crc >> 4)
        crc = self.table[(byte ^ crc) & 0xF] ^ (crc >> 4)
        self.crc = crc


class SerialPacker:
    """
    Frame sender and receiver.

    Args:
      max_idle:
        max milliseconds between bytes within a frame.
      max_packet:
        maximum packet size. Size is always transmitted as a single byte if
        max_packet is <256, otherwise two bytes are used whenever the actual
        size is >127.
      frame_start: start byte, or None for solely relying on inter-frame timeout.
      crc:
        CRC method to use. The default is our CRC16 with 0xBAAD polynomial.
    """

    # error counters
    err_crc = 0  # CRC wrong
    err_frame = 0  # length wrong or timeout

    def __init__(
        self,
        max_idle=100,
        max_packet=127,
        frame_start=FRAME_START,
        crc=CRC16,
        mark=None,
    ):
        self.max_packet = max_packet
        self.max_idle = max_idle
        self.frame_start = frame_start
        self.mark = mark
        self.mark_seen = False
        self._crc = crc
        self.reset()

    def reset(self):
        """Reset the packet receiver."""
        self.crc = self._crc()
        self.last = ticks_ms()
        self.buf = None
        self.state = 1 if self.frame_start is None else 0
        self.pos = 0
        # intentionally does not clear nbuf

    def is_idle(self):
        """
        Test whether we're currently receiving a frame.
        """
        if self.s == 0:
            return True
        t = ticks_ms()
        if ticks_diff(t, self.last) < self.max_idle:
            return False
        self.err_frame += 1
        self.reset()
        return True

    def wakeup(self):
        """
        Utility method to start a frame without a start byte. Useful if the
        first byte is used as a wake-from-sleep signal.
        """
        self.reset()
        self.s = 1

    def feed(self, byte):
        """
        Receiver for a single byte.

        Returns: a packet, if this byte completed its reception.
        """
        t = ticks_ms()
        if ticks_diff(t, self.last) >= self.max_idle:
            self.err_frame += 1
            self.reset()
        self.last = t

        s = self.state

        if self.mark is None:
            pass
        elif self.mark_seen:
            self.mark_seen = False
        elif byte == self.mark:
            self.mark_seen = True
            return None
        else:
            return byte

        if s == 0:  # idle
            if self.pos > 0:
                self.pos -= 1
                return byte
            elif byte == self.frame_start:
                s = 1
            else:
                if byte >= 0xC0:
                    # UTF-8 lead-in character. Skip 0x10xxxxxx bytes so
                    # an accidental start byte is not misrecognized.
                    c = byte ^ 0xFF
                    n = 0
                    while c:
                        n += 1
                        c >>= 1
                    self.pos = 6 - n
                return byte

        elif s == 1:  # len1
            if byte == 0:  # zero length = error (break?)
                self.err_frame += 1
                self.reset()
                return None
            self.len = byte
            if self.max_packet > 255 and (byte & 0x80):  # possibly a two byte packet
                s = 2
            elif byte > self.max_packet:
                self.err_frame += 1
                self.reset()
                return None
            else:
                s = 3
                self.buf = bytearray(self.len)
                self.pos = 0

        elif s == 2:  # len2
            self.len = (self.len & 0x7F) | (byte << 7)
            if self.len > self.max_packet:
                self.err_frame += 1
                self.reset()
                return None

            s = 3
            self.buf = bytearray(self.len)
            self.pos = 0

        elif s == 3:
            self.crc.feed(byte)
            self.buf[self.pos] = byte
            self.pos += 1
            if self.pos == self.len:
                s = 4
                self.pos = self.crc.len
                self.crc_in = 0

        elif s == 4:  # CRC
            self.crc_in = (self.crc_in << 8) | byte
            self.pos -= 1
            if self.pos == 0:
                if self.crc.crc != self.crc_in:  # error
                    self.err_crc += 1
                    rbuf = None
                else:
                    rbuf = self.buf
                self.reset()
                return rbuf

        self.state = s
        return None

    def frame(self, data):
        """
        Build framing for a message.

        Returns:
            a triple with head (start byte, length), bytes (possibly
            mark-ed), and tail (CRC) data.
        """
        ld = len(data)
        if ld > self.max_packet:
            raise ValueError(f"max len {self.max_packet}")
        crc = self._crc()
        for b in data:
            crc.feed(b)
        h = bytearray()
        if self.frame_start is not None:
            h.append(self.frame_start)
        if self.max_packet > 255 and ld > 127:
            h.extend(((ld & 0x7F) | 0x80, ld >> 7))
        else:
            h.append(ld)
        t = bytes((crc.crc >> 8, crc.crc & 0xFF))

        if self.mark is not None:

            def itermark(data):
                for b in data:
                    yield self.mark
                    yield b

            h = bytes(itermark(h))
            data = bytes(itermark(data))
            t = bytes(itermark(t))
        return h, data, t



"""
Support code to connect to a TCP server.
"""

from __future__ import annotations

import select

import asyncio

from moat.lib.micro import AC_use, log, sleep, wait_for
from moat.lib.stream import AIOBuf

_poll = select.poll()


class TcpLink(AIOBuf):
    """
    A channel that connects to a remote TCP socket.
    """

    def __init__(self, host: str, port: int, retry={}):  # noqa:B006
        self.host = host
        self.port = port
        self.retry = retry  # not changed by us
        # TODO pass the config instead of the "retry" thing

    async def stream(self):  # noqa:D102
        retry = self.retry

        async def _conn1():
            rs, ws = await asyncio.open_connection(self.host, self.port)
            if rs is not ws:
                raise RuntimeError(f"SingleSockOnly {rs!r} {ws!r}")

            # This dance is required because open_connection doesn't check
            _poll.register(rs.s, select.POLLIN)
            try:
                for x in _poll.poll(0):
                    if x[1] & (select.POLLHUP | select.POLLERR):
                        try:
                            rs.s.read(1)  # raises the applicable error
                            raise EOFError  # instead, if it's closed
                        finally:
                            rs.close()
            finally:
                _poll.unregister(rs.s)

            await AC_use(self, rs.close)
            return rs

        async def _conn():
            n = 0
            sl = retry.get("delay", 0.1)
            while True:
                try:
                    s = await _conn1()
                except OSError as e:
                    if n == 0:
                        log("Retrying: %s %d, %r", self.host, self.port, e)
                    n += 1
                    if n > retry.get("attempts", 10):
                        raise TimeoutError
                    await sleep(sl)
                    sl *= retry.get("backoff", 1.3)
                else:
                    if n:
                        log("Success: %s %d", self.host, self.port)
                    return s

        try:
            return await wait_for(retry.get("timeout", 99), _conn)
        except TimeoutError:
            log("Fail: %s %d", self.host, self.port)
            raise



"""
Message logging.
"""

from __future__ import annotations

from moat.lib.micro import AC_use, TaskGroup, idle, log
from moat.lib.rpc import BaseFwdCmd, HandlerStream, Msg, MsgSender

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from moat.util import attrdict
    from moat.lib.path import PathElem
    from moat.lib.rpc import BaseMsgHandler


class _LogStream(HandlerStream):
    other: _LogStream

    async def read_stream(self):
        """
        Stream reader.
        """
        while True:
            try:
                msg = await self.msg_out()
            except EOFError:
                self.other._log_writer.cancel()  # noqa:SLF001
                return
            await self.other.msg_in(msg)

    async def write_stream(self):
        """
        Stream writer. No-op.
        """
        async with TaskGroup() as self._log_writer:
            # taskgroup only used for its cancel scope
            await idle()
        raise EOFError


class Logger(BaseFwdCmd):
    """This is the handler for messages that forwards them to the stream."""

    doc = dict(_c=dict(_d="Subcommand logging", prefix="str:log prefix (default Msg)"))

    def __init__(self, cfg: attrdict):
        super().__init__(cfg)
        self.prefix = cfg.get("prefix", "Msg")

    @property
    def root(self):  # noqa:D102
        return self

    @property
    def sender(self):  # noqa:D102
        return self._sender

    async def setup(self):  # noqa:D102
        await super().setup()
        app = self.app
        if app is None:
            raise RuntimeError("No app")
        self.s1 = _LogStream(cast("BaseMsgHandler", self.real_root), logger=self._log)
        self.s2 = _LogStream(cast("BaseMsgHandler", app))
        self.s1.other = self.s2
        self.s2.other = self.s1
        self._sender = MsgSender(self.s2)
        await AC_use(self, self.s1)
        await AC_use(self, self.s2)

    async def handle(self, msg: Msg, rcmd: list[PathElem]) -> None:
        "Forward"
        await self.s1.handle(msg, rcmd)

    def _log(self, m: str, *a: object):
        log(f"{self.prefix}:{m}", *a)

    def cfg_reloaded(self, cfg: attrdict):
        "Forward to real root"
        self.real_root.cfg_reloaded(cfg)



"""
A hacked-up copy of some parts of `moat.util`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, TypeVar, overload

__all__ = [
    "DProxy",
    "NoProxyError",
    "Proxy",
    "as_proxy",
    "drop_proxy",
    "get_proxy",
    "name2obj",
    "obj2name",
]

NotGiven = ...  # from moat.util import NotGiven


_pkey = 1
_CProxy = {}  # name > object
_RProxy = {}  # object > name


def name2obj(name, obj=NotGiven):
    """
    Given a proxy name, return the referred object.

    If @obj is given, associate.
    """
    if obj is NotGiven and _CProxy:
        return _CProxy[name]
    _CProxy[name] = obj
    _RProxy[id(obj)] = name
    return None


def obj2name(obj):
    """
    Given a proxied object, return the name referring to it.
    """
    return _RProxy[id(obj)]


def get_proxy(obj):
    """
    Given a proxied object, return the name referring to it.

    If unknown, create a new temporary name.
    """
    try:
        return _RProxy[id(obj)]
    except KeyError:
        global _pkey
        k = "p_" + str(_pkey)
        _pkey += 1
        _CProxy[k] = obj
        _RProxy[id(obj)] = k
        return k


# def _getstate(self):
#     return (type(self), (), self.__dict__)


if TYPE_CHECKING:
    T = TypeVar("T")

    @overload
    def as_proxy(name: str) -> Callable[[T], T]: ...

    @overload
    def as_proxy(name: str, obj: Any, replace: bool = False) -> None: ...


def as_proxy(name: str, obj: Any = NotImplemented, replace: bool = False):
    """
    Export an object as a named proxy.
    Usage:

        @as_proxy("foo")
        class Foo():
            def __
    """
    # This uses NotImplemented instead of None or Ellipsis/NotGiven because
    # those two are be legitimately proxied.

    def _proxy(obj):
        "Export @obj as a proxy."
        if not replace and name in _CProxy and _CProxy[name] is not obj:
            raise ValueError("Proxy: " + repr(name) + " already exists")
        _CProxy[name] = obj
        _RProxy[id(obj)] = name
        #       if isinstance(obj,type) and not hasattr(obj,"__getstate__"):
        #           obj.__getstate__ = _getstate
        return obj

    if obj is NotImplemented:
        return _proxy
    else:
        _proxy(obj)
        return obj


def drop_proxy(p):
    """
    After sending a proxy we keep it in memory in case the remote returns
    it, or an expression with it.

    If that won't happen, the remote needs to tell us to clean it up.
    """
    if not isinstance(p, str):
        p = _RProxy[id(p)]
    if p == "" or p[0] == "_":
        raise ValueError("Can't delete a system proxy")
    r = _CProxy.pop(p)
    del _RProxy[id(r)]


class NoProxyError(ValueError):
    "Error for nonexistent proxy values"

    # pylint:disable=unnecessary-pass


class Proxy:
    """
    A proxy object, i.e. a placeholder for things that cannot pass
    through a codec. No object data are included.
    """

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name!r})"

    def ref(self):
        """Dereferences the proxy"""
        return name2obj(self.name)


class DProxy(Proxy):
    """
    A proxy object with data. This is implemented as a type that's proxied,
    thus the object can be reconstituted by the receiver (if it knows the
    proxy class) or at least rebuilt when the original sender gets the
    proxy structure back (if it doesn't). The object's state is included.
    """

    def __init__(self, name, a, k):
        super().__init__(name)
        self.a = a
        self.k = k

    def __getitem__(self, i):
        if i in self.k:
            return self.k[i]
        else:
            try:
                return self.a[i]
            except TypeError:
                from moat.lib.micro import log  # noqa: PLC0415

                log("*ERR %r", self.k)
                raise KeyError(i) from None

    def __eq__(self, other):
        if not isinstance(other, DProxy):
            return NotImplemented

        # Split into several lines so we can selectively set breakpoints
        # when debugging
        if self.name != other.name:
            return False
        if self.a != other.a or self.k != other.k:
            return False
        return True

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.name!r},"
            + ",".join(repr(x) for x in (self.a, self.k))
            + ")"
        )

    def ref(self):
        """Dereferences the proxy"""
        return name2obj(self.name)
